/**
 * SMSCore.java
 *
 *
 * @author Copyright (C) 2002-2003 Chris White
 * @version 18th January 2003
 * @see "JavaGear Final Project Report"
 */

/*
   This file is part of JavaGear.

   JavaGear is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   JavaGear is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with JavaGear; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


import java.io.*;
import com.nttdocomo.ui.Graphics;

public final class SMSCore extends Thread implements Runnable {
	private final SMSCanvas gui;
	private final SMSApu audio;
	private final PSGi psgi;

	// Z80CPU REGISTERS
	// Accumulator
	private int A_  = 0;
	private int A2_ = 0;
	// Flag Register
	private int F_  = 0;
	private int F2_ = 0;
	// Loop Counter
	private int B_  = 0;
	private int B2_ = 0;
	private int C_  = 0;
	private int C2_ = 0;
	// General Purpose
	private int D_  = 0;
	private int D2_ = 0;
	private int E_  = 0;
	private int E2_ = 0;
	// General Purpose
	private int H_  = 0;
	private int H2_ = 0;
	private int L_  = 0;
	private int L2_ = 0;
	// Index Page Address Register
	private int IX_ = 0;
	private int IY_ = 0;

	private int pc = 0x0000;						// Program Counter
	private int sp = 0x0000;						// Stack Pointer
	private boolean iff1 = false;						// Interrupt Flip Flop 1
	private boolean iff2 = false;						// Interrupt Flip Flop 2
	private boolean halt	= false;					// Halt Instruction
	private boolean EI_inst = false;					// Previous Instruction was EI
	private boolean irq = false;						// Z80 Interrupt Line
	private int R_ = 0;							// Memory Refresh Register
	private int I_ = 0;							// Interrupt page address register
	private int tstates = 0;						// T States

//*	Setup									// General Parameters
	private int is_sms = 0;							// SMS Mode
	private boolean pause_button = false;					// SMS Pause Button
	private int h_start = 0;						// Horizontal Viewport Start
	private int h_end   = 256;						// Horizontal Viewport End
	private int v_start = 0;						// Vertical Viewport Start
	private int v_end   = 192;						// Vertical Viewport End

//*	Port									// Port Emulation
	private int nat1 = 0;							// Nationalisation Bit 1
	private int nat2 = 0;							// Nationalisation Bit 2

//*	Vdp									// VDP Emulation
	private int[] display = new int[0xC100];				// Pointer to Current Display
	private int[] display_zoom = new int[0xCA80];				// Pointer to Current Display
	private byte[] VRAM = new byte[0x4000];					// Video RAM
	private int[] CRAM = new int[0x40];					// Colour RAM
	private int vdpreg[] = new int[0x10];					// VDP Registers
	private int status = 0;							// Status Register
	private boolean first_byte = true;					// First or Second Byte of Command Word
	private int command_byte = 0;						// Command Word First Byte Latch
	private int location_VRAM;						// Location in VRAM
	private int operation;							// Store type of operation taking place
	private int read_buffer;						// Buffer VRAM Reads
	private int line;							// Current Line Number
	private int counter;							// Vertical Line Interrupt Counter
	private boolean lineint;						// Line interrupt Pending
	private boolean frameint;						// Frame interrupt Pending
	private boolean bg_priority[] = new boolean[0x107];			// Background Priorites
	private boolean spritecol[] = new boolean[0x100];			// Sprite Collisions

//*     Control
	public static final int REQUEST_START = 0;
	public static final int REQUEST_RESET = 1;
	public static final int REQUEST_LOAD = 2;
	public static final int REQUEST_SAVE = 3;
	public static final int REQUEST_SLEEP = 4;
	private int request;
	private int param;

//*	Memory
	private byte[] mem = new byte[0x10000];					// Array of Main Memory
	private byte [][] pages  = new byte[4][0x4000];				// Array of Memory Pages
	private byte [][] cart_ram_pages = new byte[2][0x4000];			// Array of Cartridge RAM Pages
	private boolean frame_two_rom = false;					// Stores whether frame 2 contains cartridge ROM or cartridge RAM
	private int cart_ram_page = 0;						// Stores which Page of cartridge RAM to use
	private int number_of_pages = 4;					// Stores number of pages
	private int[] daa;

	private boolean i_sw = false;
	private boolean emu_running = false;
	private int cycles = 0xFFFFFFFF;
	private byte[] romCheck = new byte[0x10];
	private final byte[] ShiningForceGaiden = {-13,-19,86,-61,104,0,0,0,-13,-41,-5,-55,0,0,0,0};
	private int dummyInt = 0;
	private boolean dummyBool = false;
	private final int stateVer = 46;

																				// ------ OPCODE TIMINGS -----
	private final short opstates[] = {
										/*          0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F */
										/* 0x00 */  4,10, 7, 6, 4, 4, 7, 4, 4,11, 7, 6, 4, 4, 7, 4,
										/* 0x10 */  8,10, 7, 6, 4, 4, 7, 4,12,11, 7, 6, 4, 4, 7, 4,
										/* 0x20 */  7,10,16, 6, 4, 4, 7, 4, 7,11,16, 6, 4, 4, 7, 4,
										/* 0x30 */  7,10,13, 6,11,11,10, 4, 7,11,13, 6, 4, 4, 7, 4,
										/* 0x40 */  4, 4, 4, 4, 4, 4, 7, 4, 4, 4, 4, 4, 4, 4, 7, 4,
										/* 0x50 */  4, 4, 4, 4, 4, 4, 7, 4, 4, 4, 4, 4, 4, 4, 7, 4,
										/* 0x60 */  4, 4, 4, 4, 4, 4, 7, 4, 4, 4, 4, 4, 4, 4, 7, 4,
										/* 0x70 */  7, 7, 7, 7, 7, 7, 4, 7, 4, 4, 4, 4, 4, 4, 7, 4,
										/* 0x80 */  4, 4, 4, 4, 4, 4, 7, 4, 4, 4, 4, 4, 4, 4, 7, 4,
										/* 0x90 */  4, 4, 4, 4, 4, 4, 7, 4, 4, 4, 4, 4, 4, 4, 7, 4,
										/* 0xA0 */  4, 4, 4, 4, 4, 4, 7, 4, 4, 4, 4, 4, 4, 4, 7, 4,
										/* 0xB0 */  4, 4, 4, 4, 4, 4, 7, 4, 4, 4, 4, 4, 4, 4, 7, 4,
										/* 0xC0 */  5,10,10,10,10,11, 7,11, 5,10,10, 0,10,17, 7,11,
										/* 0xD0 */  5,10,10,11,10,11, 7,11, 5, 4,10,11,10, 0, 7,11,
										/* 0xE0 */  5,10,10,19,10,11, 7,11, 5, 4,10, 4,10, 0, 7,11,
										/* 0xF0 */  5,10,10, 4,10,11, 7,11, 5, 6,10, 4,10, 0, 7,11 };
	private final short opcbstates[] = {
										/*          0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F */
										/* 0x00 */  8, 8, 8, 8, 8, 8,15, 8, 8, 8, 8, 8, 8, 8,15, 8,
										/* 0x10 */  8, 8, 8, 8, 8, 8,15, 8, 8, 8, 8, 8, 8, 8,15, 8,
										/* 0x20 */  8, 8, 8, 8, 8, 8,15, 8, 8, 8, 8, 8, 8, 8,15, 8,
										/* 0x30 */  8, 8, 8, 8, 8, 8,15, 8, 8, 8, 8, 8, 8, 8,15, 8,
										/* 0x40 */  8, 8, 8, 8, 8, 8,12, 8, 8, 8, 8, 8, 8, 8,12, 8,
										/* 0x50 */  8, 8, 8, 8, 8, 8,12, 8, 8, 8, 8, 8, 8, 8,12, 8,
										/* 0x60 */  8, 8, 8, 8, 8, 8,12, 8, 8, 8, 8, 8, 8, 8,12, 8,
										/* 0x70 */  8, 8, 8, 8, 8, 8,12, 8, 8, 8, 8, 8, 8, 8,12, 8,
										/* 0x80 */  8, 8, 8, 8, 8, 8,15, 8, 8, 8, 8, 8, 8, 8,15, 8,
										/* 0x90 */  8, 8, 8, 8, 8, 8,15, 8, 8, 8, 8, 8, 8, 8,15, 8,
										/* 0xA0 */  8, 8, 8, 8, 8, 8,15, 8, 8, 8, 8, 8, 8, 8,15, 8,
										/* 0xB0 */  8, 8, 8, 8, 8, 8,15, 8, 8, 8, 8, 8, 8, 8,15, 8,
										/* 0xC0 */  8, 8, 8, 8, 8, 8,15, 8, 8, 8, 8, 8, 8, 8,15, 8,
										/* 0xD0 */  8, 8, 8, 8, 8, 8,15, 8, 8, 8, 8, 8, 8, 8,15, 8,
										/* 0xE0 */  8, 8, 8, 8, 8, 8,15, 8, 8, 8, 8, 8, 8, 8,15, 8,
										/* 0xF0 */  8, 8, 8, 8, 8, 8,15, 8, 8, 8, 8, 8, 8, 8,15, 8 };
	private final short opddstates[] = {
										/*          0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F */
										/* 0x00 */  0, 0, 0, 0, 0, 0, 0, 0, 0,15, 0, 0, 0, 0, 0, 0,
										/* 0x10 */  0, 0, 0, 0, 0, 0, 0, 0, 0,15, 0, 0, 0, 0, 0, 0,
										/* 0x20 */  0,14,20,10, 9, 9, 9, 0, 0,15,20,10, 9, 9, 9, 0,
										/* 0x30 */  0, 0, 0, 0,23,23,19, 0, 0,15, 0, 0, 0, 0, 0, 0,
										/* 0x40 */  0, 0, 0, 0, 9, 9,19, 0, 0, 0, 0, 0, 9, 9,19, 0,
										/* 0x50 */  0, 0, 0, 0, 9, 9,19, 0, 0, 0, 0, 0, 9, 9,19, 0,
										/* 0x60 */  9, 9, 9, 9, 9, 9,19, 9, 9, 9, 9, 9, 9, 9,19, 9,
										/* 0x70 */ 19,19,19,19,19,19,19,19, 0, 0, 0, 0, 9, 9,19, 0,
										/* 0x80 */  0, 0, 0, 0, 9, 9,19, 0, 0, 0, 0, 0, 9, 9,19, 0,
										/* 0x90 */  0, 0, 0, 0, 9, 9,19, 0, 0, 0, 0, 0, 9, 9,19, 0,
										/* 0xA0 */  0, 0, 0, 0, 9, 9,19, 0, 0, 0, 0, 0, 9, 9,19, 0,
										/* 0xB0 */  0, 0, 0, 0, 9, 9,19, 0, 0, 0, 0, 0, 9, 9,19, 0,
										/* 0xC0 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
										/* 0xD0 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
										/* 0xE0 */  0,14, 0,23, 0,15, 0, 0, 0, 8, 0, 0, 0, 0, 0, 0,
										/* 0xF0 */  0, 0, 0, 0, 0, 0, 0, 0, 0,10, 0, 0, 0, 0, 0, 0};
	private final short opedstates[] = {
										/*          0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F */
										/* 0x00 */  8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
										/* 0x10 */  8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
										/* 0x20 */  8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
										/* 0x30 */  8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
										/* 0x40 */ 12,12,15,20, 8,14, 8, 9,12,12,15,20, 8,14, 8, 9,
										/* 0x50 */ 12,12,15,20, 8,14, 8, 9,12,12,15,20, 8,14, 8, 9,
										/* 0x60 */ 12,12,15,20, 8,14, 8,18,12,12,15,20, 8,14, 8,18,
										/* 0x70 */ 12,12,15,20, 8,14, 8, 8,12,12,15,20, 8,14, 8, 8,
										/* 0x80 */  8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
										/* 0x90 */  8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
										/* 0xA0 */ 16,16,16,16, 8, 8, 8, 8,16,16,16,16, 8, 8, 8, 8,
										/* 0xB0 */  0, 0, 0, 0, 8, 8, 8, 8, 0, 0, 0, 0, 8, 8, 8, 8,
										/* 0xC0 */  8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
										/* 0xD0 */  8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
										/* 0xE0 */  8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
										/* 0xF0 */  8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8 };
	private final boolean PARITY[] = {
										/*		   0      1      2      3      4      5      6      7      8      9      A      B      C      D      E      F   */
										/* 0x00 */ true,  false, false, true,  false, true,  true,  false, false, true,  true,  false, true,  false, false, true, 
										/* 0x10 */ false, true,  true,  false, true,  false, false, true,  true,  false, false, true,  false, true,  true,  false,
										/* 0x20 */ false, true,  true,  false, true,  false, false, true,  true,  false, false, true,  false, true,  true,  false,
										/* 0x30 */ true,  false, false, true,  false, true,  true,  false, false, true,  true,  false, true,  false, false, true, 
										/* 0x40 */ false, true,  true,  false, true,  false, false, true,  true,  false, false, true,  false, true,  true,  false,
										/* 0x50 */ true,  false, false, true,  false, true,  true,  false, false, true,  true,  false, true,  false, false, true, 
										/* 0x60 */ true,  false, false, true,  false, true,  true,  false, false, true,  true,  false, true,  false, false, true, 
										/* 0x70 */ false, true,  true,  false, true,  false, false, true,  true,  false, false, true,  false, true,  true,  false,
										/* 0x80 */ false, true,  true,  false, true,  false, false, true,  true,  false, false, true,  false, true,  true,  false,
										/* 0x90 */ true,  false, false, true,  false, true,  true,  false, false, true,  true,  false, true,  false, false, true, 
										/* 0xA0 */ true,  false, false, true,  false, true,  true,  false, false, true,  true,  false, true,  false, false, true, 
										/* 0xB0 */ false, true,  true,  false, true,  false, false, true,  true,  false, false, true,  false, true,  true,  false,
										/* 0xC0 */ true,  false, false, true,  false, true,  true,  false, false, true,  true,  false, true,  false, false, true, 
										/* 0xD0 */ false, true,  true,  false, true,  false, false, true,  true,  false, false, true,  false, true,  true,  false,
										/* 0xE0 */ false, true,  true,  false, true,  false, false, true,  true,  false, false, true,  false, true,  true,  false,
										/* 0xF0 */ true,  false, false, true,  false, true,  true,  false, false, true,  true,  false, true,  false, false, true};


/*
	// DoJaG~pɎc
	private final int SMS[] = {-16777216, -11206656, -5636096, -65536, -16755456, -11184896, -5614336, -43776, -16733696, -11163136, -5592576, -22016, -16711936,
									-11141376, -5570816, -256, -16777131, -11206571, -5636011, -65451, -16755371, -11184811, -5614251, -43691, -16733611, -11163051,
									-5592491, -21931, -16711851, -11141291, -5570731, -171, -16777046, -11206486, -5635926, -65366, -16755286, -11184726, -5614166,
									-43606, -16733526, -11162966, -5592406, -21846, -16711766, -11141206, -5570646, -86, -16776961, -11206401, -5635841, -65281,
									-16755201, -11184641, -5614081, -43521, -16733441, -11162881, -5592321, -21761, -16711681, -11141121, -5570561, -1
								};
	private final int GG1[] = {-16777216, -15728640, -14680064, -13631488, -12582912, -11534336, -10485760, -9437184, -8388608, -7340032, -6291456, -5242880,
									-4194304, -3145728, -2097152, -1048576, -16773120, -15724544, -14675968, -13627392, -12578816, -11530240, -10481664, -9433088,
									-8384512, -7335936, -6287360, -5238784, -4190208, -3141632, -2093056, -1044480, -16769024, -15720448, -14671872, -13623296,
									-12574720, -11526144, -10477568, -9428992, -8380416, -7331840, -6283264, -5234688, -4186112, -3137536, -2088960, -1040384,
									-16764928, -15716352, -14667776, -13619200, -12570624, -11522048, -10473472, -9424896, -8376320, -7327744, -6279168, -5230592,
									-4182016, -3133440, -2084864, -1036288, -16760832, -15712256, -14663680, -13615104, -12566528, -11517952, -10469376, -9420800,
									-8372224, -7323648, -6275072, -5226496, -4177920, -3129344, -2080768, -1032192, -16756736, -15708160, -14659584, -13611008,
									-12562432, -11513856, -10465280, -9416704, -8368128, -7319552, -6270976, -5222400, -4173824, -3125248, -2076672, -1028096,
									-16752640, -15704064, -14655488, -13606912, -12558336, -11509760, -10461184, -9412608, -8364032, -7315456, -6266880, -5218304,
									-4169728, -3121152, -2072576, -1024000, -16748544, -15699968, -14651392, -13602816, -12554240, -11505664, -10457088, -9408512,
									-8359936, -7311360, -6262784, -5214208, -4165632, -3117056, -2068480, -1019904, -16744448, -15695872, -14647296, -13598720,
									-12550144, -11501568, -10452992, -9404416, -8355840, -7307264, -6258688, -5210112, -4161536, -3112960, -2064384, -1015808,
									-16740352, -15691776, -14643200, -13594624, -12546048, -11497472, -10448896, -9400320, -8351744, -7303168, -6254592, -5206016,
									-4157440, -3108864, -2060288, -1011712, -16736256, -15687680, -14639104, -13590528, -12541952, -11493376, -10444800, -9396224,
									-8347648, -7299072, -6250496, -5201920, -4153344, -3104768, -2056192, -1007616, -16732160, -15683584, -14635008, -13586432,
									-12537856, -11489280, -10440704, -9392128, -8343552, -7294976, -6246400, -5197824, -4149248, -3100672, -2052096, -1003520,
									-16728064, -15679488, -14630912, -13582336, -12533760, -11485184, -10436608, -9388032, -8339456, -7290880, -6242304, -5193728,
									-4145152, -3096576, -2048000, -999424, -16723968, -15675392, -14626816, -13578240, -12529664, -11481088, -10432512, -9383936,
									-8335360, -7286784, -6238208, -5189632, -4141056, -3092480, -2043904, -995328, -16719872, -15671296, -14622720, -13574144,
									-12525568, -11476992, -10428416, -9379840, -8331264, -7282688, -6234112, -5185536, -4136960, -3088384, -2039808, -991232,
									-16715776, -15667200, -14618624, -13570048, -12521472, -11472896, -10424320, -9375744, -8327168, -7278592, -6230016, -5181440,
									-4132864, -3084288, -2035712, -987136
								};
	private final int GG2[] = {0, 16, 32, 48, 64, 80, 96, 112, 128, 144, 160, 176, 192, 208, 224, 240};
	private void setColor() {
		int j;
		int c;
		for (j = 0; j < SMS_JAVA.length; j++) {
			c = SMS[j];
			SMS_JAVA[j] = Graphics.getColorOfRGB((c >> 16) & 0xff, (c >> 8) & 0xff, c & 0xff);
		}
		for (j = 0; j < GG_JAVA1.length; j++) {
			c = GG1[j];
			GG_JAVA1[j] = Graphics.getColorOfRGB((c >> 16) & 0xff, (c >> 8) & 0xff, c & 0xff);
		}
		for (j = 0; j < GG_JAVA2.length; j++) {
			c = GG2[j];
			GG_JAVA2[j] = Graphics.getColorOfRGB((c >> 16) & 0xff, (c >> 8) & 0xff, c & 0xff);
		}
	}
*/



	public SMSCore(SMSCanvas gui_, PSGi psgi_) {
		gui = gui_;
		audio = new SMSApu(psgi_);
		psgi = psgi_;
		daa();
		// DoJaG~
//		if(System.getProperty("microedition.platform").startsWith("ap"))setColor();
	}



	public final void reset() {
		F_   = 0;
		F2_  = 0;
		A_   = 0;
		A2_  = 0;
		B_   = 0;
		B2_  = 0;
		C_   = 0;
		C2_  = 0;
		D_   = 0;
		D2_  = 0;
		E_   = 0;
		E2_  = 0;
		H_   = 0;
		H2_  = 0;
		L_   = 0;
		L2_  = 0;
		IX_  = 0;
		IY_  = 0;
		R_   = 0;
		I_   = 0;
		pc   = 0x0000;
		sp   = 0xDFF0;
		tstates = 0;
		//IM_ = 0;
		iff1 = false;
		iff2 = false;
		EI_inst = false;
		halt = false;

		cart_ram_page = 0;
		frame_two_rom = true;
		if (pages != null) {
			System.arraycopy(pages[0],0x0000,mem,0x0000,0x4000);
			System.arraycopy(pages[1],0x0000,mem,0x4000,0x4000);
		}
		first_byte = true;
		command_byte = 0;
		lineint = false;
		frameint = false;
		location_VRAM = 0;
		counter = 0;
		status = 0;
		operation = 0;
		vdpreg[0] = 0;
		vdpreg[1] = 0;
		vdpreg[2] = 0x0E;	   // B1-B3 high on startup
		vdpreg[3] = 0;
		vdpreg[4] = 0;
		vdpreg[5] = 0x7E;	   // B1-B6 high on startup
		vdpreg[6] = 0;
		vdpreg[7] = 0;
		vdpreg[8] = 0;
		vdpreg[9] = 0;
		vdpreg[10] = 0;

		// Clear Memory
		for (int i = display.length; i-- != 0;)
			display[i] = 0;
		for (int i = VRAM.length; i-- != 0;)
			VRAM[i] = 0;
		for (int i = CRAM.length; i-- != 0;)
			CRAM[i] = 0;
		for (int i = bg_priority.length; i-- != 0;)
			bg_priority[i] = false;
		for (int i = spritecol.length; i-- != 0;)
			spritecol[i] = false;
		irq = false;
		switch (is_sms) {
			case 0:
				h_start = 0;
				h_end   = 256;
				v_start = 0;
				v_end   = 192;
				break;
			case 1:
				h_start = 40;
				h_end   = 160 + 48;
				v_start = 24;
				v_end   = 144 + 24;
				break;
		}
		audio.init();
	}



	private final void daa()
		{
			daa = new int[0x800];
			final int NEGATIVE = 0x100;
			final int CARRY = 0x200;
			final int HALFCARRY = 0x400;
			for (int x = 0; x < 0x800; x++)
			{
				int a = x & 0xFF;
				int highNibble = (a >> 4);
				int lowNibble  = (a & 0x0F);
				boolean carry = false;
				if ((x & NEGATIVE) == 0) // ADD, ADC, INC
				{
					if ((x & CARRY) == 0)
					{
						if ((x & HALFCARRY) == 0)
						{
							if ((highNibble <= 0x08) && (lowNibble <= 0x0F) && (lowNibble >= 0x0A)) // Line 2
								a+=0x06;
							else if ((highNibble <= 0x09) && (lowNibble <= 0x09)); // Line 1 of Page 236
							else if ((highNibble >= 0x0A) && (lowNibble <= 0x09))  // Line 4
							{
								a+=0x60;
								carry = true;
							}
							else if ((highNibble >= 0x09) && (lowNibble >= 0x0A)) // Line 5
							{
								a+=0x66;
								carry = true;
							}
						}
						else // Half Carry is 1, Carry is 0, Negative is 0
						{
							if ((highNibble <= 0x09) && (lowNibble <= 0x03)) // Line 3
								a+=0x06;
							else if ((highNibble >= 0x0A) && (lowNibble <= 0x03)) // Line 6
							{
								a+=0x66;
								carry = true;
							}
						}
					}
					else // Flag Carry is 1
					{
						if ((x & HALFCARRY) == 0) // C = 1, HC = 0, N = 0
						{
							if ((highNibble <= 0x02) && (lowNibble <= 0x09)) // Line 7
							{
								a+=0x60;
								carry = true;
							}
							else if ((highNibble <= 0x02) && (lowNibble >= 0x0A)) // Line 8
							{
								a+=0x66;
								carry = true;
							}
						}
						else // Half Carry is 1
						{
							if ((highNibble < 0x04) && (lowNibble < 0x04)) // Line 9
							{
								a+=0x66;
								carry = true;
							}
						}
					}
				}
				// Negative Flag is true SUB, SBC, DEC, NEG --------------------------------------
				else
				{
					if ((x & CARRY) == 0)
					{
						if ((x & HALFCARRY) == 0)
						{
							if ((highNibble < 0x0A) && (lowNibble < 0x0A) );  // Line 10
						}
						else // Flag HalfCarry is 1, Carry = 0, Neg = 1
						{
							if ((highNibble < 0x09) && (lowNibble > 0x05)) // Line 11
								a +=0xFA;
						}
					}
					else // Flag Carry is 1
					{
						if ((x & HALFCARRY) == 0)
						{
							if ((highNibble >= 0x07) && (lowNibble <= 0x09)) // Line 12
							{
								a+=0xA0;
								carry = true;
							}
						}
						else // Flag HalfCarry is 1
						{
							if ((highNibble >= 0x06) && (lowNibble >= 0x06)) // Line 13
							{
								a +=0x9A;
								carry = true;
							}
						}
					}
				}
				a &= 0xFF;
				if (carry) a|=0x100;
				daa[x] = a;
			} // end for
		}



	/**
	 *  Read a cartridge into memory
	 *
	 *  @param is
	 *  @param size
	 *  @param extension
	 *
	 *  @return			 True if cartridge read successfully
	 */

	public final int loadROM(InputStream is, long size) throws IOException {
		if (size < 16384) {
			return 0;
		}
		try {
			// Strip 512 Byte File Headers
			if ((size % 1024) != 0) {
				is.skip(512); // skip 512 bytes
				size -=512;
			}
			// Calculate number of pages from file size and create array appropriately
			number_of_pages = (int) (size/0x4000);
			pages = new byte[number_of_pages][0x4000];
			// Read file into pages array
			for (int x = 0; x < number_of_pages; x++) {
				// second value is offset, third is length
				is.read(pages[x], 0x0000, 0x4000);
			}
			// Default Mapping (Needed or Shinobi doesn't work)
			System.arraycopy(pages[0],0x0000,mem,0x0000,0x4000);
			System.arraycopy(pages[1],0x0000,mem,0x4000,0x4000);
			for (int i = 0; i < 0x10; i++)
				romCheck[i] = pages[0][i];
			boolean normal = false;
			for (int i = 0; i < romCheck.length; i++)
				if (romCheck[i] != ShiningForceGaiden[i])
					normal = true;
			if (normal) cycles = 227;
			else cycles = 322;
		} catch(Exception e) {
			return 0;
		}
		//System.out.println("Lo :" + mem[0x7FFA] + " Hi :" + mem[0x7FFB]);
		if(mem[0x7FFA] != 0 || mem[0x7FFB] != 0) return 1;
		else return 2;
	}



	private final void mem_write(int adr, int val) {
		if (adr < 0xC000) {
			// Cart RAM
			if (adr >= 0x8000) {
				if (!frame_two_rom)
					mem[adr] = cart_ram_pages[cart_ram_page][adr&0x3FFF] = (byte) val;
			}
			return;
		}
		// Program RAM
		else if (adr < 0xE000) {
			mem[adr] = mem[adr+0x2000] = (byte) val;
			return;
		}
		else {
			if (adr >= 0xFFFC) {
				switch (adr) {
					// RAM/ROM select register
					case 0xFFFC:
						// Map RAM
						if ((val & 0x8) == 0x8) {
							frame_two_rom = false;
							if ((val & 0x4) == 0x4) {
								System.arraycopy(cart_ram_pages[1],0,mem,0x8000,0x4000);
								cart_ram_page = 1;
							}
							else {
								System.arraycopy(cart_ram_pages[0],0,mem,0x8000,0x4000);
								cart_ram_page = 0;
							}
						}
						// Map ROM
						else
							frame_two_rom = true;
						break;
					// Page 0 ROM Bank
					case 0xFFFD:
						System.arraycopy(pages[val%number_of_pages],0,mem,0x0000,0x4000);
						break;
					// Page 1 ROM Bank
					case 0xFFFE:
						System.arraycopy(pages[val%number_of_pages],0,mem,0x4000,0x4000);
						break;
					// Page 2 ROM Bank
					case 0xFFFF:
						if (frame_two_rom)
							System.arraycopy(pages[val%number_of_pages],0,mem,0x8000,0x4000);
						else
							System.arraycopy(pages[cart_ram_page],0,mem,0x8000,0x4000);
						break;
				}
			}
			// Mirror RAM
			mem[adr] = mem[adr&0xDFFF] = (byte) val;
			return;
		}
	}



	private final int  mem_read(int adr) {
		return mem[adr]&0xFF;
	}



	private final int mem_readSigned(int adr) {
		return mem[adr];
	}



        public final void request(int _request, int _param) {
            request = _request;
            param = _param;
        }



	public final void run() {
		while (emu_running) {
                    switch (request) {
                    case 1 :
                        reset();
                        break;

                    case 2 :
                        {
                            SDbinding sd = null;
                            DataInputStream dis = null;
                            try {
                                sd = new SDbinding();
                                dis = sd.openDataInputStream(gui.getSaveFile(param));
                                if (dis != null){
                                    stateLoad(dis);
                                }
                            } catch (IOException e) {
                            } finally {
                                sd.close();
                                dis = null;
                                sd = null;
                            }
                        }
                        break;

                    case 3 :
                        {
                            SDbinding sd = null;
                            DataOutputStream dos = null;
                            try {
                                sd = new SDbinding();
                                dos = sd.openDataOutputStream(gui.getSaveFile(param));
                                if (dos != null){
                                    stateSave(dos);
                                    dos.flush();
                                }
                            } catch (IOException e) {
                            } finally {
                                sd.close();
                                dos = null;
                                sd = null;
                            }
                        }
                        break;

                    case 4 :
                        continue;

                    default :
                        break;
                    }
                    request = 0;
			final int CYCLES_PER_LINE = i_sw ? cycles * 53 / 100 : cycles;
			boolean draw_frame = gui.startFrame();
			for (line = 0; line < 262; line++) {										//drawFrame(lineno,draw_frame);
				if (line < 192) {
					if (line == 191) {
						status |= 0x80;
						frameint = true;
					}
					if (counter == 0) {
						counter = vdpreg[10];
						lineint = true;
					}
					else counter--;
					if (lineint && ((vdpreg[0] & 0x10) == 0x10))
						irq = true;
				} else {
					counter = vdpreg[10];
					if (frameint && ((vdpreg[1] & 0x20) == 0x20) && (line < 224))
						irq = true;
				}
																						// Draw Next Line
				if (draw_frame && v_start <= line && line < v_end) {					//drawLine(lineno);
					if ((vdpreg[1] & 0x40) != 0) {
						// bg
						int lineno_bg = line;
						if (vdpreg[9] != 0)
							lineno_bg += vdpreg[9] & 0x07;
						int pattern = 0;
						int pal = 0;
						int hscroll = 0;
						int start_column = 0;
						int start_row = lineno_bg;
						int tilex = 0;
						int tiley = (lineno_bg & 7);
						if (((vdpreg[0] & 0x40) == 0x40) && (line < 16) && is_sms == 0)
							hscroll = 0;
						else if (vdpreg[8] != 0) {
							hscroll = vdpreg[8] & 0x07;
							start_column = 32 - (vdpreg[8]>>3);
						}
						if (is_sms == 1)
							start_column = (start_column + 5) & 0x1F;
						if (vdpreg[9] != 0) {
							start_row += vdpreg[9] & 0xF8;
							if (start_row > 223) start_row -= 224;
						}
						for (int x = h_start; x < h_end; x += 8) {
							if (start_column == 32) start_column = 0;
							if (((vdpreg[0] & 0x80) == 0x80) && (x > 184)) {
								start_row = line;
								tiley = (line & 7);
							}
							int tile_props = ((vdpreg[2] & 0x0f &~ 0x01) << 10) + (start_column << 1) + ((start_row - tiley) << 3);
							int firstbyte =  VRAM[tile_props]&0xFF;
							int secondbyte = VRAM[tile_props+1]&0xFF;
							int priority = secondbyte & 0x10;
							if ((secondbyte & 0x08) == 0x08) pal = 16;
							else pal = 0;
							pattern = (firstbyte + ((secondbyte & 0x01)<<8))<<5;
							int address;
							if ((secondbyte & 0x04) == 0)
								address = (tiley<<2) + pattern;
							else
								address = ((7-tiley)<<2)+pattern;
							tilex = 0;
							int xpos;
							int row_precal = line << 8;
							int address0 = VRAM[address+0];
							int address1 = VRAM[address+1];
							int address2 = VRAM[address+2];
							int address3 = VRAM[address+3];
							for(int bit = 0x80; bit != 0; bit >>= 1) {
								if ((secondbyte & 0x02) == 0)
									xpos = tilex + hscroll + x;
								else
									xpos = 7 - tilex + hscroll + x;
								int colour = 0;
								if ((address0 & bit) != 0)
									colour |= 0x01;
								if ((address1 & bit) != 0)
									colour |= 0x02;
								if ((address2 & bit) != 0)
									colour |= 0x04;
								if ((address3 & bit) != 0)
									colour |= 0x08;
								if ((priority == 0x10) && (colour != 0))
									bg_priority[xpos] = true;
								else
									bg_priority[xpos] = false;
								display[xpos+row_precal] = CRAM[colour+pal];
								tilex++;
							}
							start_column++;
						}
						// end bg

						// sprite
						spritecol = new boolean[0x100];
						int sat = (vdpreg[5] &~ 0x01 &~ 0x80)<<7;
						int count = 0;
						int height = 8;
						boolean zoomed = false;
						if ((vdpreg[1] & 0x02) == 0x02)
							height = 16;
						if ((vdpreg[1] & 0x01) == 0x01) {
							height <<= 1;
							zoomed = true;
						}
						for (int spriteno = 0; spriteno < 0x40; spriteno++) {
							if (count >= 8) {
								status |= 0x40;
								break;
							}
							int y = VRAM[sat + spriteno]&0xFF;
							int address = sat+(spriteno<<1);
							int x = VRAM[address + 0x80]&0xFF;
							int i = VRAM[address + 0x81]&0xFF;
							if (y == 208) break;
							y++;
							if ((vdpreg[6] & 0x04) == 0x04)
								i |= 0x100;
							if ((vdpreg[1] & 0x02) == 0x02)
								i &= ~0x01;
							if ((vdpreg[0] & 0x08) == 0x08)
								x -= 8;
							if (y > 240)
								y -= 256;
							if ((line >= y) && ((line-y) < height)) {
								int row_precal = line << 8;
								if (!zoomed) {
									int adr = (i<<5) + ((line-y)<<2);
									for(int bit = 0x80; bit != 0; bit>>=1) {
										if ((x >= 0) && (y >= 0) && (x <= 255)) {
											plotSpritePixel(x, x+row_precal, adr, bit);
										}
										x++;
									}
								} else {
									int adr = (i<<5) + ((line-y)<<1);
									int pixel = 0;
									for(int bit = 0x80; bit != 0; bit>>=1) {
										if ((x+pixel+1 >= 0) && (y >= 0) && (x <= 255)) {
											plotSpritePixel(x+pixel, x+pixel+row_precal, adr, bit);
											plotSpritePixel(x+pixel+1, x+pixel+1+row_precal, adr, bit);
										}
										x++;
										pixel++;
									}
								}
								count++;
							}
						}
						// end sprite
					} else {
						for (int x = 255; x >= 0; x--)
							display[x + (line << 8)] = CRAM[16 + (vdpreg[7] & 0x0F)];
					}
					if (is_sms == 0 && (vdpreg[0] & 0x20) == 0x20) {
						for (int x = 7; x > 0; x--)
							display[x + (line << 8)] = CRAM[16 + (vdpreg[7] & 0x0F)];
					}
				} // end drawLine
				// emulate Z80
				Z80(CYCLES_PER_LINE);
			} // end for

			if (pause_button) {
				iff2 = iff1;
				R_ = (R_&0x80) | ((R_+1)&0x7F);
				// If we're in a halt instruction, increment the PC and get out of it
				if (halt) {
					pc++;
					halt = false;
				}
				mem_write(--sp, pc >> 8);mem_write(--sp, pc & 0xFF);			// Preserve PC on stack
				pc = 0x66;
				tstates -= 11;
				pause_button = iff1 = false;
			}
			// Refresh Java Screenpanel
			if (draw_frame) {
				if (is_sms == 1 && gui.getZoom() != 0) drawFrame();
				gui.endFrame();
			}
			// Output Sound
			audio.doFrame();
		} //end while()
	} // end run()



	private final void plotSpritePixel(int x, int plotlocation_VRAM, int address, int bit){
		int colour = 0;
		if ((VRAM[address] & bit) != 0) colour |= 1;
		if ((VRAM[address+1] & bit) != 0) colour |= 0x02;
		if ((VRAM[address+2] & bit) != 0) colour |= 0x04;
		if ((VRAM[address+3] & bit) != 0) colour |= 0x08;
		if (!(bg_priority[x]) && (colour != 0)){
			if (!spritecol[x]){
				spritecol[x] = true;
				display[plotlocation_VRAM] = CRAM[colour+16];
			} else {
				status |= 0x20; // set bit 5 of status flag
			}
		}
	}



	private final void Z80(int CYCLES_PER_LINE)
	{
		int opcode = 0;
		tstates = CYCLES_PER_LINE;
		while (tstates > 0) {
			if (irq) { // Check for interrupt
				if (iff1 && !EI_inst) {;
					if (halt) {
						pc++;
						halt = false;
					}
					R_ = (R_&0x80) | ((R_+1)&0x7F);
					iff1 = false;
					iff2 = false;
					irq = false;
					//if ((IM_ == 1) || (IM_ == 0)) {
						mem_write(--sp, pc >> 8);
						mem_write(--sp, pc & 0xFF);		// Preserve PC on stack
						pc = 0x38;
						tstates -= 13;
					//}
				}
			}

			opcode = (int)mem[pc]&0xFF;				// Fetch Opcode
			tstates -= opstates[opcode];		// Decrement TStates
			//doOp(opcode);						// Interpret Opcode
			{
				R_++;
				EI_inst = false;
				switch(opcode) {
					case 0x00:// NOP
						pc++;
						break;
					case 0x01:// LD BC,nn
						C_  = mem_read(pc+1);
						B_  = mem_read(pc+2);
						pc+=3;
						break;
					case 0x02:// LD (BC),A
						mem_write((B_ << 8) | C_, A_);
						pc++;
						break;
					case 0x03:// INC BC
						C_ = (C_ + 1) & 0xFF;
						if (C_ == 0) B_ = (B_ + 1) & 0xFF;
						pc++;
						break;
					case 0x04:// INC B
						B_ = inc8(B_);
						pc++;
						break;
					case 0x05:// DEC B
						B_ = dec8(B_);
						pc++;
						break;
					case 0x06:// LD B,n
						B_ = mem_read(pc+1);
						pc+=2;
						break;
					case 0x07:// RLCA
						{
							int temp = A_;
							A_ = (A_ << 1) & 0xFF;
							if ((temp & 0x80) == 0x80){
								F_ |= 0x01;
								A_ |= 0x01;
							} else {
								F_ &=~ 0x01;
							}
							F_ &=~ 0x02; F_ &=~ 0x10;
						}
						pc++;
						break;
					case 0x08://EX AF AF'
						{
							int temp;
							temp = F_;
							F_ = F2_;
							F2_ = temp;
							temp = A_;
							A_ = A2_;
							A2_ = temp;
						}
						pc++;
						break;
					case 0x09:// ADD HL,BC
						{
							int value = ((B_ << 8) | C_);
							int reg = ((H_ << 8) | L_);
							int result = reg + value;
							int carrybits = reg ^ value ^ result;
							if ((carrybits & 0x1000) != 0 ) F_ |= 0x10;
							else F_ &=~ 0x10;
							if ((carrybits & 0x10000) != 0) F_ |= 0x01;
							else F_ &=~ 0x01;
							result &= 0xFFFF;
							L_   = result & 0xFF;
							H_  = (result >> 8) & 0xFF;
							F_ &=~ 0x02;
							result >>= 8;			// Bits 3 and 5 affected by H_ byte addition
							if((result & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;}
							if((result & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;}
						}
						pc++;
						break;
					case 0x0A:// LD A,(BC)
						A_ = mem_read((B_ << 8) | C_);
						pc++;
						break;
					case 0x0B:// DEC BC
						C_ = (C_ - 1) & 0xFF;
						if (C_ == 255) B_ = (B_ - 1) & 0xFF;
						pc++;
						break;
					case 0x0C:// INC C
						C_ = inc8(C_);
						pc++;
						break;
					case 0x0D:// DEC C
						C_ = dec8(C_);
						pc++;
						break;
					case 0x0E:// LD C,n
						C_ = mem_read(pc+1);
						pc+=2;
						break;
					case 0x0F:// RRCA
						{
							int temp = A_;
							A_ >>= 1;
							if ((temp & 0x01) != 0) {
								F_ |= 0x01;
								A_ |= 0x80;
							} else {
								F_ &=~ 0x01;
							}
							F_ &=~ 0x02; F_ &=~ 0x10;
						}
						pc++;
						break;
					case 0x10:// DJNZ (PC+e)
						{
							int temp = (B_-1)&0xff;
							B_ = (temp) & 0xFF;
							if (temp != 0) {
								pc+= mem_readSigned(pc+1) +2;
								tstates -= 5;
							}
							else pc+=2;
						}
						break;
					case 0x11:// LD DE,nn
						E_ = mem_read(pc + 1);
						D_ = mem_read(pc + 2);
						pc+=3;
						break;
					case 0x12:// LD (DE), A
						mem_write(((D_ << 8) | E_), A_);
						pc++;
						break;
					case 0x13:// INC DE
						E_ = (E_ + 1) & 0xFF;
						if (E_ == 0) D_ = (D_ + 1) & 0xFF;
						pc++;
						break;
					case 0x14:// INC D
						D_ = inc8(D_);
						pc++;
						break;
					case 0x15:// DEC D
						D_ = dec8(D_);
						pc++;
						break;
					case 0x16:// LD D,n
						D_ = mem_read(pc+1);
						pc+=2;
						break;
					case 0x17:// RLA
						{
							int temp = A_;
							A_ = (A_ << 1) & 0xFF;
							if ((F_ & 0x01) == 0x01) A_ |= 0x01;
							if ((temp & 0x80) == 0x80) F_ |= 0x01;
							else F_ &=~ 0x01;
							F_ &=~ 0x02; F_ &=~ 0x10;
						}
						pc++;
						break;
					case 0x18:// JR (PC+e)
						pc+= mem_readSigned(pc+1) + 2;
						break;
					case 0x19:// ADD HL,DE
						{
							int value = ((D_ << 8) | E_);
							int reg = ((H_ << 8) | L_);
							int result = reg + value;
							int carrybits = reg ^ value ^ result;
							if ((carrybits & 0x1000) != 0 ) F_ |= 0x10;
							else F_ &=~ 0x10;
							if ((carrybits & 0x10000) != 0) F_ |= 0x01;
							else F_ &=~ 0x01;
							result &= 0xFFFF;
							L_   = result & 0xFF;
							H_  = (result >> 8) & 0xFF;
							F_ &=~ 0x02;
							result >>= 8;			// Bits 3 and 5 affected by H_ byte addition
							if((result & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;}
							if((result & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;}
						}
						pc++;
						break;
					case 0x1A:// LD A,(DE)
						A_ = mem_read((D_ << 8) | E_);
						pc++;
						break;
					case 0x1B:// DEC DE
						E_ = (E_ - 1) & 0xFF;
						if (E_ == 0xFF) D_ = (D_ - 1) & 0xFF;
						pc++;
						break;
					case 0x1C:// INC E
						E_ = inc8(E_);
						pc++;
						break;
					case 0x1D:// DEC E
						E_ = dec8(E_);
						pc++;
						break;
					case 0x1E:// LD E,N
						E_ = mem_read(pc+1);
						pc+=2;
						break;
					case 0x1F:// RRA (c)
						{
							int temp = A_;
							A_ >>= 1;
							if ((F_ & 0x01) == 0x01) A_ |= 0x80;
							if ((temp & 0x01) != 0) F_ |= 0x01;
							else F_ &=~ 0x01;
							F_ &=~ 0x02; F_ &=~ 0x10;
						}
						pc++;
						break;
					case 0x20:// JR NZ,(PC+e)
						if ((F_ & 0x40) != 0x40) {
							pc+= mem_readSigned(pc+1) +2;
							tstates -= 5;
						}
						else pc+=2;
						break;
					case 0x21:// LD HL,nn
						L_ = mem_read(pc + 1);
						H_ = mem_read(pc + 2);
						pc+=3;
						break;
					case 0x22:// LD (nn),HL
						mem_write((mem_read(pc + 2) << 8) + mem_read(pc + 1), L_);
						mem_write((mem_read(pc + 2) << 8) + mem_read(pc + 1) + 1, H_);
						pc+=3;
						break;
					case 0x23:// INC HL
						L_ = (L_ + 1) & 0xFF;
						if (L_ == 0) H_ = (H_ + 1) & 0xFF;
						pc++;
						break;
					case 0x24:// INC H
						H_ = inc8(H_);
						pc++;
						break;
					case 0x25:// DEC H
						H_ = dec8(H_);
						pc++;
						break;
					case 0x26:// LD H,n
						H_ = mem_read(pc+1);
						pc+=2;
						break;
					case 0x27:// DAA
						{
							if ((F_ & 0x02) == 0x02)
								A_ |= 0x100;
							if ((F_ & 0x01) == 0x01)
								A_ |= 0x200;
							if ((F_ & 0x10) == 0x10)
								A_ |= 0x400;
							int temp = daa[A_];
							if (temp > 0xFF) F_ |= 0x01;
							else F_ &=~ 0x01;
							temp &= 0xFF;
							A_ = temp;
							if(PARITY[temp]){F_ |= 0x04;}else{F_ &=~ 0x04;};
							if (temp == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							if ((temp & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
						}
						pc++;
						break;
					case 0x28:// JR Z,(PC+e)
						if ((F_ & 0x40) == 0x40) {
							pc+= mem_readSigned(pc+1) +2;
							tstates -= 5;
						}
						else pc+=2;
						break;
					case 0x29:// ADD HL,HL
						{
							int value = ((H_ << 8) | L_);
							int reg = ((H_ << 8) | L_);
							int result = reg + value;
							int carrybits = reg ^ value ^ result;
							if ((carrybits & 0x1000) != 0 ) F_ |= 0x10;
							else F_ &=~ 0x10;
							if ((carrybits & 0x10000) != 0) F_ |= 0x01;
							else F_ &=~ 0x01;
							result &= 0xFFFF;
							L_   = result & 0xFF;
							H_  = (result >> 8) & 0xFF;
							F_ &=~ 0x02;
							result >>= 8;			// Bits 3 and 5 affected by H_ byte addition
							if((result & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;}
							if((result & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;}
						}
						pc++;
						break;
					case 0x2A:// LD HL,(nn)
						L_ = mem_read((mem_read(pc + 2) << 8) + mem_read(pc + 1));
						H_ = mem_read((mem_read(pc + 2) << 8) + mem_read(pc + 1) + 1);
						pc+=3;
						break;
					case 0x2B:// DEC HL
						L_ = (L_ - 1) & 0xFF;
						if (L_ == 255) H_ = (H_ - 1) & 0xFF;
						pc++;
						break;
					case 0x2C:// INC L
						L_ = inc8(L_);
						pc++;
						break;
					case 0x2D:// DEC L
						L_ = dec8(L_);
						pc++;
						break;
					case 0x2E:// LD L,n
						L_ = mem_read(pc+1);
						pc+=2;
						break;
					case 0x2F:// CPL
						{
							A_ =(~A_) & 0xFF;
							F_ |= 0x02;
							F_ |= 0x10;
							if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0x30:// JR NC,(PC+e)
						if ((F_ & 0x01) != 0x01) {
							pc+= mem_readSigned(pc+1) +2;
							tstates -= 5;
						}
						else pc+=2;
						break;
					case 0x31:// LD SP,nn
						sp = (mem_read(pc + 2) << 8) + mem_read(pc + 1);
						pc+=3;
						break;
					case 0x32:// LD (nn),A
						mem_write((mem_read(pc + 2) << 8) + mem_read(pc + 1), A_);
						pc+=3;
						break;
					case 0x33:// INC SP
						sp++;
						pc++;
						break;
					case 0x34:// INC (HL)
						inchl(((H_ << 8) | L_));
						pc++;
						break;
					case 0x35:// DEC (HL)
						dechl(((H_ << 8) | L_));
						pc++;
						break;
					case 0x36:// LD (HL),n
						mem_write(((H_ << 8) | L_), mem_read(pc+1));
						pc+=2;
						break;
					case 0x37:// SCF
						F_ |= 0x01;
						F_ &=~ 0x02;
						F_ &=~ 0x10;
						pc++;
						break;
					case 0x38:// JR C,(PC+e)
						if ((F_ & 0x01) == 0x01) {
							pc+= mem_readSigned(pc+1) +2;
							tstates -=5 ;
						}
						else pc+=2;
						break;
					case 0x39:// ADD HL,SP
						{
							int value = sp;
							int reg = ((H_ << 8) | L_);
							int result = reg + value;
							int carrybits = reg ^ value ^ result;
							if ((carrybits & 0x1000) != 0 ) F_ |= 0x10;
							else F_ &=~ 0x10;
							if ((carrybits & 0x10000) != 0) F_ |= 0x01;
							else F_ &=~ 0x01;
							result &= 0xFFFF;
							L_   = result & 0xFF;
							H_  = (result >> 8) & 0xFF;
							F_ &=~ 0x02;
							result >>= 8;			// Bits 3 and 5 affected by H_ byte addition
							if((result & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;}
							if((result & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;}
						}
						pc++;
						break;
					case 0x3A:// LD A,(nn)
						A_ = mem_read((mem_read(pc + 2) << 8) + mem_read(pc + 1));
						pc+=3;
						break;
					case 0x3B:// DEC SP
						sp--;
						pc++;
						break;
					case 0x3C:// INC A
						A_ = (A_ + 1) & 0xFF;
						if ((A_ & 0x0F) == 0x00)
							F_ |= 0x10;
						else
							F_ &=~ 0x10;
						F_ &=~ 0x02;
						if (A_ == 0x80) F_ |= 0x04;
						else F_ &=~ 0x04;
						if (A_ == 0) F_ |= 0x40;
						else F_ &=~ 0x40;
						if ((A_ & 0x80) == 0) F_ &=~ 0x80;
						else F_ |= 0x80;
						pc++;
						break;
					case 0x3D:// DEC A
						A_ = (A_ - 1) & 0xFF;
						if ((A_ & 0x0F) == 0x0F)
							F_ |= 0x10;
						else
							F_ &=~ 0x10;
						F_ |= 0x02;
						if (A_ == 0x7F) F_ |= 0x04;
						else F_ &=~ 0x04;
						if (A_ == 0) F_ |= 0x40;
						else F_ &=~ 0x40;
						if ((A_ & 0x80) == 0) F_ &=~ 0x80;
						else F_ |= 0x80;
						pc++;
						break;
					case 0x3E:// LD A,n
						A_ = mem_read(pc+1);
						pc+=2;
						break;
					case 0x3F:// CCF
						if (((F_ & 0x01) == 0x01)) {
							F_ &=~ 0x01;
							F_ |= 0x10;
						} else {
							F_ |= 0x01;
							F_ &=~ 0x10;
						}
						F_ &=~ 0x02;
						pc++;
						break;
					case 0x40:											pc++; break;		// LD B,B
					case 0x41: B_ = C_;									pc++; break;		// LD B,C
					case 0x42: B_ = D_;									pc++; break;		// LD B,D
					case 0x43: B_ = E_;									pc++; break;		// LD B,E
					case 0x44: B_ = H_;									pc++; break;		// LD B,H
					case 0x45: B_ = L_;									pc++; break;		// LD B,L
					case 0x46: B_ = mem_read((H_ << 8) | L_);			pc++; break;		// LD B,(HL)
					case 0x47: B_ = A_;									pc++; break;		// LD B,A
					case 0x48: C_ = B_;									pc++; break;		// LD C,B
					case 0x49:											pc++; break;		// LD C,C
					case 0x4A: C_ = D_;									pc++; break;		// LD C,D
					case 0x4B: C_ = E_;									pc++; break;		// LD C,E
					case 0x4C: C_ = H_;									pc++; break;		// LD C,H
					case 0x4D: C_ = L_;									pc++; break;		// LD C,L
					case 0x4E: C_ = mem_read((H_ << 8) | L_);			pc++; break;		// LD C,(HL)
					case 0x4F: C_ = A_;									pc++; break;		// LD C,A
					case 0x50: D_ = B_;									pc++; break;		// LD D,B
					case 0x51: D_ = C_;									pc++; break;		// LD D,C
					case 0x52:											pc++; break;		// LD D,D
					case 0x53: D_ = E_;									pc++; break;		// LD D,E
					case 0x54: D_ = H_;									pc++; break;		// LD D,H
					case 0x55: D_ = L_;									pc++; break;		// LD D,L
					case 0x56: D_ = mem_read((H_ << 8) | L_);			pc++; break;		// LD D,(HL)
					case 0x57: D_ = A_;									pc++; break;		// LD D,A
					case 0x58: E_ = B_;									pc++; break;		// LD E,B
					case 0x59: E_ = C_;									pc++; break;		// LD E,C
					case 0x5A: E_ = D_;									pc++; break;		// LD E,D
					case 0x5B:											pc++; break;		// LD E,E
					case 0x5C: E_ = H_;									pc++; break;		// LD E,H
					case 0x5D: E_ = L_;									pc++; break;		// LD E,L
					case 0x5E: E_ = mem_read((H_ << 8) | L_);			pc++; break;		// LD E,(HL)
					case 0x5F: E_ = A_;									pc++; break;		// LD E,A
					case 0x60: H_ = B_;									pc++; break;		// LD H,B
					case 0x61: H_ = C_;									pc++; break;		// LD H,C
					case 0x62: H_ = D_;									pc++; break;		// LD H,D
					case 0x63: H_ = E_;									pc++; break;		// LD H,E
					case 0x64:											pc++; break;		// LD H,H
					case 0x65: H_ = L_;									pc++; break;		// LD H,L
					case 0x66: H_ = mem_read((H_ << 8) | L_);			pc++; break;		// LD H,(HL)
					case 0x67: H_ = A_;									pc++; break;		// LD H,A
					case 0x68: L_ = B_;									pc++; break;		// LD L,B
					case 0x69: L_ = C_;									pc++; break;		// LD L,C
					case 0x6A: L_ = D_;									pc++; break;		// LD L,D
					case 0x6B: L_ = E_;									pc++; break;		// LD L,E
					case 0x6C: L_ = H_;									pc++; break;		// LD L,H
					case 0x6D:											pc++; break;		// LD L,L
					case 0x6E: L_ = mem_read((H_ << 8) | L_);			pc++; break;		// LD L,(HL)
					case 0x6F: L_ = A_;									pc++; break;		// LD L,A
					case 0x70: mem_write(((H_ << 8) | L_), B_); 		pc++; break;		// LD (HL),B
					case 0x71: mem_write(((H_ << 8) | L_), C_); 		pc++; break;		// LD (HL),C
					case 0x72: mem_write(((H_ << 8) | L_), D_); 		pc++; break;		// LD (HL),D
					case 0x73: mem_write(((H_ << 8) | L_), E_); 		pc++; break;		// LD (HL),E
					case 0x74: mem_write(((H_ << 8) | L_), H_); 		pc++; break;		// LD (HL),H
					case 0x75: mem_write(((H_ << 8) | L_), L_); 		pc++; break;		// LD (HL),L
					case 0x76: halt = true;									  break;		// HALT
					case 0x77: mem_write(((H_ << 8) | L_), A_);			pc++; break;		// LD (HL),A
					case 0x78: A_ = B_;									pc++; break;		// LD A,B
					case 0x79: A_ = C_;									pc++; break;		// LD A,C
					case 0x7A: A_ = D_;									pc++; break;		// LD A,D
					case 0x7B: A_ = E_;									pc++; break;		// LD A,E
					case 0x7C: A_ = H_;									pc++; break;		// LD A,H
					case 0x7D: A_ = L_;									pc++; break;		// LD A,L
					case 0x7E: A_ = mem_read((H_ << 8) | L_);			pc++; break;		// LD A,(HL)
					case 0x7F:											pc++; break;		// LD A,A
					case 0x80: A_ = add(A_,B_);							pc++; break;		// ADD A,B
					case 0x81: A_ = add(A_,C_);							pc++; break;		// ADD A,C
					case 0x82: A_ = add(A_,D_);							pc++; break;		// ADD A,D
					case 0x83: A_ = add(A_,E_);							pc++; break;		// ADD A,E
					case 0x84: A_ = add(A_,H_);							pc++; break;		// ADD A,H
					case 0x85: A_ = add(A_,L_);							pc++; break;		// ADD A,L
					case 0x86: A_ = add(A_,mem_read((H_ << 8) | L_));	pc++; break;		// ADD A,(HL)
					case 0x87: A_ = add(A_,A_);							pc++; break;		// ADD A,A
					case 0x88: A_ = adc(A_,B_);							pc++; break;		// ADC A,B
					case 0x89: A_ = adc(A_,C_);							pc++; break;		// ADC A,C
					case 0x8A: A_ = adc(A_,D_);							pc++; break;		// ADC A,D
					case 0x8B: A_ = adc(A_,E_);							pc++; break;		// ADC A,E
					case 0x8C: A_ = adc(A_,H_);							pc++; break;		// ADC A,H
					case 0x8D: A_ = adc(A_,L_);							pc++; break;		// ADC A,L
					case 0x8E: A_ = adc(A_,mem_read((H_ << 8) | L_));	pc++; break;		// ADC A,(HL)
					case 0x8F: A_ = adc(A_,A_);							pc++; break;		// ADC A,A
					case 0x90: A_ = sub(A_,B_);							pc++; break;		// SUB A,B
					case 0x91: A_ = sub(A_,C_);							pc++; break;		// SUB A,C
					case 0x92: A_ = sub(A_,D_);							pc++; break;		// SUB A,D
					case 0x93: A_ = sub(A_,E_);							pc++; break;		// SUB A,E
					case 0x94: A_ = sub(A_,H_);							pc++; break;		// SUB A,H
					case 0x95: A_ = sub(A_,L_);							pc++; break;		// SUB A,L
					case 0x96: A_ = sub(A_,mem_read((H_ << 8) | L_));	pc++; break;		// SUB A,(HL)
					case 0x97: A_ = sub(A_,A_);							pc++; break;		// SUB A,A
					case 0x98: A_ = sbc(A_,B_);							pc++; break;		// SBC A,B
					case 0x99: A_ = sbc(A_,C_);							pc++; break;		// SBC A,C
					case 0x9A: A_ = sbc(A_,D_);							pc++; break;		// SBC A,D
					case 0x9B: A_ = sbc(A_,E_);							pc++; break;		// SBC A,E
					case 0x9C: A_ = sbc(A_,H_);							pc++; break;		// SBC A,H
					case 0x9D: A_ = sbc(A_,L_);							pc++; break;		// SBC A,L
					case 0x9E: A_ = sbc(A_,mem_read((H_ << 8) | L_));	pc++; break;		// SBC A,(HL)
					case 0x9F: A_ = sbc(A_,A_);							pc++; break;		// SBC A,A
					case 0xA0:// AND A,B
						{
							int value = B_;
							A_ &= value;
							if ((A_ & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							if (A_ == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							F_ |= 0x10;
							if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
							F_ &=~ 0x02;
							F_ &=~ 0x01;
							if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xA1:// AND A,C
						{
							int value = C_;
							A_ &= value;
							if ((A_ & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							if (A_ == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							F_ |= 0x10;
							if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
							F_ &=~ 0x02;
							F_ &=~ 0x01;
							if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xA2:// AND A,D
						{
							int value = D_;
							A_ &= value;
							if ((A_ & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							if (A_ == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							F_ |= 0x10;
							if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
							F_ &=~ 0x02;
							F_ &=~ 0x01;
							if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xA3:// AND A,E
						{
							int value = E_;
							A_ &= value;
							if ((A_ & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							if (A_ == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							F_ |= 0x10;
							if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
							F_ &=~ 0x02;
							F_ &=~ 0x01;
							if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xA4:// AND A,H
						{
							int value = H_;
							A_ &= value;
							if ((A_ & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							if (A_ == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							F_ |= 0x10;
							if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
							F_ &=~ 0x02;
							F_ &=~ 0x01;
							if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xA5:// AND A,L
						{
							int value = L_;
							A_ &= value;
							if ((A_ & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							if (A_ == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							F_ |= 0x10;
							if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
							F_ &=~ 0x02;
							F_ &=~ 0x01;
							if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xA6:// AND A,(HL)
						{
							int value = mem_read((H_ << 8) | L_);
							A_ &= value;
							if ((A_ & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							if (A_ == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							F_ |= 0x10;
							if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
							F_ &=~ 0x02;
							F_ &=~ 0x01;
							if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xA7:// AND A,A
						{
							int value = A_;
							A_ &= value;
							if ((A_ & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							if (A_ == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							F_ |= 0x10;
							if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
							F_ &=~ 0x02;
							F_ &=~ 0x01;
							if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xA8:// XOR A,B
						{
							int value = B_;
							A_ ^= value; // bitwise exclusive OR
							if (A_ == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							if ((A_ & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							F_ &=~ 0x10;
							F_ &=~ 0x02;
							F_ &=~ 0x01;
							if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
							if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xA9:// XOR A,C
						{
							int value = C_;
							A_ ^= value; // bitwise exclusive OR
							if (A_ == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							if ((A_ & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							F_ &=~ 0x10;
							F_ &=~ 0x02;
							F_ &=~ 0x01;
							if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
							if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xAA:// XOR A,D
						{
							int value = D_;
							A_ ^= value; // bitwise exclusive OR
							if (A_ == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							if ((A_ & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							F_ &=~ 0x10;
							F_ &=~ 0x02;
							F_ &=~ 0x01;
							if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
							if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xAB:// XOR A,E
						{
							int value = E_;
							A_ ^= value; // bitwise exclusive OR
							if (A_ == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							if ((A_ & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							F_ &=~ 0x10;
							F_ &=~ 0x02;
							F_ &=~ 0x01;
							if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
							if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xAC:// XOR A,H
						{
							int value = H_;
							A_ ^= value; // bitwise exclusive OR
							if (A_ == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							if ((A_ & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							F_ &=~ 0x10;
							F_ &=~ 0x02;
							F_ &=~ 0x01;
							if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
							if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xAD:// XOR A,L
						{
							int value = L_;
							A_ ^= value; // bitwise exclusive OR
							if (A_ == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							if ((A_ & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							F_ &=~ 0x10;
							F_ &=~ 0x02;
							F_ &=~ 0x01;
							if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
							if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xAE:// XOR A,(HL)
						{
							int value = mem_read((H_ << 8) | L_);
							A_ ^= value; // bitwise exclusive OR
							if (A_ == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							if ((A_ & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							F_ &=~ 0x10;
							F_ &=~ 0x02;
							F_ &=~ 0x01;
							if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
							if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xAF:// XOR A,A
						{
							int value = A_;
							A_ ^= value; // bitwise exclusive OR
							if (A_ == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							if ((A_ & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							F_ &=~ 0x10;
							F_ &=~ 0x02;
							F_ &=~ 0x01;
							if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
							if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xB0:// OR A,B
						{
							int value = B_;
							A_ |= value; // bitwise inclusive OR to turn relevant bits on
							if ((A_ & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							if (A_ == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							F_ &=~ 0x10;
							if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
							F_ &=~ 0x02;
							F_ &=~ 0x01;
							if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xB1:// OR A,C
						{
							int value = C_;
							A_ |= value; // bitwise inclusive OR to turn relevant bits on
							if ((A_ & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							if (A_ == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							F_ &=~ 0x10;
							if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
							F_ &=~ 0x02;
							F_ &=~ 0x01;
							if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xB2:// OR A,D
						{
							int value = D_;
							A_ |= value; // bitwise inclusive OR to turn relevant bits on
							if ((A_ & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							if (A_ == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							F_ &=~ 0x10;
							if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
							F_ &=~ 0x02;
							F_ &=~ 0x01;
							if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xB3:// OR A,E
						{
							int value = E_;
							A_ |= value; // bitwise inclusive OR to turn relevant bits on
							if ((A_ & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							if (A_ == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							F_ &=~ 0x10;
							if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
							F_ &=~ 0x02;
							F_ &=~ 0x01;
							if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xB4:// OR A,H
						{
							int value = H_;
							A_ |= value; // bitwise inclusive OR to turn relevant bits on
							if ((A_ & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							if (A_ == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							F_ &=~ 0x10;
							if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
							F_ &=~ 0x02;
							F_ &=~ 0x01;
							if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xB5:// OR A,L
						{
							int value = L_;
							A_ |= value; // bitwise inclusive OR to turn relevant bits on
							if ((A_ & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							if (A_ == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							F_ &=~ 0x10;
							if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
							F_ &=~ 0x02;
							F_ &=~ 0x01;
							if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xB6:// OR A,(HL)
						{
							int value = mem_read((H_ << 8) | L_);
							A_ |= value; // bitwise inclusive OR to turn relevant bits on
							if ((A_ & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							if (A_ == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							F_ &=~ 0x10;
							if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
							F_ &=~ 0x02;
							F_ &=~ 0x01;
							if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xB7:// OR A,A
						{
							int value = A_;
							A_ |= value; // bitwise inclusive OR to turn relevant bits on
							if ((A_ & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							if (A_ == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							F_ &=~ 0x10;
							if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
							F_ &=~ 0x02;
							F_ &=~ 0x01;
							if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xB8:// CP A,B
						{
							int value = B_;
							int result = A_ - value;
							int carrybits = A_ ^ value ^ result;
							if ((carrybits & 0x10) != 0 ) F_ |= 0x10;
							else F_ &=~ 0x10;
							if ((carrybits & 0x100) != 0) F_ |= 0x01;
							else F_ &=~ 0x01;
							if ((((carrybits << 1) ^ carrybits) & 0x100) != 0) F_ |= 0x04;
							else F_ &=~ 0x04;
							result &= 0xFF;
							if ((result & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							if (result == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							F_ |= 0x02;
							if((value & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((value & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xB9:// CP A,C
						{
							int value = C_;
							int result = A_ - value;
							int carrybits = A_ ^ value ^ result;
							if ((carrybits & 0x10) != 0 ) F_ |= 0x10;
							else F_ &=~ 0x10;
							if ((carrybits & 0x100) != 0) F_ |= 0x01;
							else F_ &=~ 0x01;
							if ((((carrybits << 1) ^ carrybits) & 0x100) != 0) F_ |= 0x04;
							else F_ &=~ 0x04;
							result &= 0xFF;
							if ((result & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							if (result == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							F_ |= 0x02;
							if((value & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((value & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xBA:// CP A,D
						{
							int value = D_;
							int result = A_ - value;
							int carrybits = A_ ^ value ^ result;
							if ((carrybits & 0x10) != 0 ) F_ |= 0x10;
							else F_ &=~ 0x10;
							if ((carrybits & 0x100) != 0) F_ |= 0x01;
							else F_ &=~ 0x01;
							if ((((carrybits << 1) ^ carrybits) & 0x100) != 0) F_ |= 0x04;
							else F_ &=~ 0x04;
							result &= 0xFF;
							if ((result & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							if (result == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							F_ |= 0x02;
							if((value & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((value & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xBB:// CP A,E
						{
							int value = E_;
							int result = A_ - value;
							int carrybits = A_ ^ value ^ result;
							if ((carrybits & 0x10) != 0 ) F_ |= 0x10;
							else F_ &=~ 0x10;
							if ((carrybits & 0x100) != 0) F_ |= 0x01;
							else F_ &=~ 0x01;
							if ((((carrybits << 1) ^ carrybits) & 0x100) != 0) F_ |= 0x04;
							else F_ &=~ 0x04;
							result &= 0xFF;
							if ((result & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							if (result == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							F_ |= 0x02;
							if((value & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((value & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xBC:// CP A,H
						{
							int value = H_;
							int result = A_ - value;
							int carrybits = A_ ^ value ^ result;
							if ((carrybits & 0x10) != 0 ) F_ |= 0x10;
							else F_ &=~ 0x10;
							if ((carrybits & 0x100) != 0) F_ |= 0x01;
							else F_ &=~ 0x01;
							if ((((carrybits << 1) ^ carrybits) & 0x100) != 0) F_ |= 0x04;
							else F_ &=~ 0x04;
							result &= 0xFF;
							if ((result & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							if (result == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							F_ |= 0x02;
							if((value & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((value & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xBD:// CP A,L
						{
							int value = L_;
							int result = A_ - value;
							int carrybits = A_ ^ value ^ result;
							if ((carrybits & 0x10) != 0 ) F_ |= 0x10;
							else F_ &=~ 0x10;
							if ((carrybits & 0x100) != 0) F_ |= 0x01;
							else F_ &=~ 0x01;
							if ((((carrybits << 1) ^ carrybits) & 0x100) != 0) F_ |= 0x04;
							else F_ &=~ 0x04;
							result &= 0xFF;
							if ((result & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							if (result == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							F_ |= 0x02;
							if((value & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((value & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xBE:// CP A,(HL)
						{
							int value = mem_read((H_ << 8) | L_);
							int result = A_ - value;
							int carrybits = A_ ^ value ^ result;
							if ((carrybits & 0x10) != 0 ) F_ |= 0x10;
							else F_ &=~ 0x10;
							if ((carrybits & 0x100) != 0) F_ |= 0x01;
							else F_ &=~ 0x01;
							if ((((carrybits << 1) ^ carrybits) & 0x100) != 0) F_ |= 0x04;
							else F_ &=~ 0x04;
							result &= 0xFF;
							if ((result & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							if (result == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							F_ |= 0x02;
							if((value & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((value & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xBF:// CP A,A
						{
							int value = A_;
							int result = A_ - value;
							int carrybits = A_ ^ value ^ result;
							if ((carrybits & 0x10) != 0 ) F_ |= 0x10;
							else F_ &=~ 0x10;
							if ((carrybits & 0x100) != 0) F_ |= 0x01;
							else F_ &=~ 0x01;
							if ((((carrybits << 1) ^ carrybits) & 0x100) != 0) F_ |= 0x04;
							else F_ &=~ 0x04;
							result &= 0xFF;
							if ((result & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							if (result == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							F_ |= 0x02;
							if((value & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((value & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xC0:// RET NZ
						if ((F_ & 0x40) != 0x40) {
							pc = (mem_read(sp + 1) << 8) + mem_read(sp);
							sp += 2;
							sp &= 0xFFFF;
							tstates -= 6;
						}
						else pc++;
						break;
					case 0xC1:// POP BC
						C_ = mem_read(sp);
						B_ = mem_read(sp + 1);
						sp += 2;
						sp &= 0xFFFF;
						pc++;
						break;
					case 0xC2:// JP NZ,(nn)
						if ((F_ & 0x40) != 0x40) pc = (mem_read(pc + 2) << 8) + mem_read(pc + 1);
						else pc+=3;
						break;
					case 0xC3:// JP (nn)
						pc = (mem_read(pc + 2) << 8) + mem_read(pc + 1);
						break;
					case 0xC4:// CALL NZ (nn)
						if ((F_ & 0x40) != 0x40) {
							mem_write(--sp, (pc + 3) >> 8);
							mem_write(--sp, (pc + 3) & 0xFF);
							pc = (mem_read(pc + 2) << 8) + mem_read(pc + 1);
							tstates -= 7;
						}
						else pc += 3;
						break;
					case 0xC5:// PUSH BC
						mem_write(--sp, B_);
						mem_write(--sp, C_);
						pc++;
						break;
					case 0xC6:// ADD A,n
						A_ = add(A_,mem_read(++pc));
						pc++;
						break;
					case 0xC7:// RST 00H
						mem_write(--sp, (pc + 1) >> 8);
						mem_write(--sp, (pc + 1) & 0xFF);
						pc = 0;
						break;
					case 0xC8:// RET Z
						if ((F_ & 0x40) == 0x40) {
							pc = (mem_read(sp+ 1) << 8) + mem_read(sp);
							sp += 2;
							sp &= 0xFFFF;
							tstates -= 6;
						}
						else pc++;
						break;
					case 0xC9:// RET
						pc = (mem_read(sp + 1) << 8) + mem_read(sp);
						sp += 2;
						sp &= 0xFFFF;
						break;
					case 0xCA:// JP Z,(nn)
						if ((F_ & 0x40) == 0x40) pc = (mem_read(pc + 2) << 8) + mem_read(pc + 1);
						else pc+=3;
						break;
					case 0xCB:// CB Opcode
						doCB(mem_read(++pc));
						break;
					case 0xCC:// CALL Z (nn)
						if ((F_ & 0x40) == 0x40) {
							mem_write(--sp, (pc + 3) >> 8);
							mem_write(--sp, (pc + 3) & 0xFF);
							pc = (mem_read(pc + 2) << 8) + mem_read(pc + 1);
							tstates -= 7;
						}
						else pc += 3;
						break;
					case 0xCD:// CALL (nn)
						mem_write(--sp, (pc + 3) >> 8);
						mem_write(--sp, (pc + 3) & 0xFF);
						pc = (mem_read(pc + 2) << 8) + mem_read(pc + 1);
						break;
					case 0xCE:// ADC A,n
						A_ = adc(A_,mem_read(++pc));
						pc++;
						break;
					case 0xCF:// RST 08H
						mem_write(--sp, (pc + 1) >> 8);
						mem_write(--sp, (pc + 1) & 0xFF);
						pc = 8;
						break;
					case 0xD0:// RET NC
						if ((F_ & 0x01) != 0x01) {
							pc = (mem_read(sp + 1) << 8) + mem_read(sp);
							sp += 2;
							sp &= 0xFFFF;
							tstates -= 6;
						}
						else pc++;
						break;
					case 0xD1:// POP DE
						E_ = mem_read(sp);
						D_ = mem_read(sp + 1);
						sp += 2;
						sp &= 0xFFFF;
						pc++;
						break;
					case 0xD2:// JP NC,(nn)
						if ((F_ & 0x01) != 0x01) pc = (mem_read(pc + 2) << 8) + mem_read(pc + 1);
						else pc+=3;
						break;
					case 0xD3:// OUT (n),A
						out(mem_read(++pc), A_);
						pc++;
						break;
					case 0xD4:// CALL NC (nn)
						if ((F_ & 0x01) != 0x01) {
							mem_write(--sp, (pc + 3) >> 8);
							mem_write(--sp, (pc + 3) & 0xFF);
							pc = (mem_read(pc + 2) << 8) + mem_read(pc + 1);
							tstates -= 7;
						}
						else pc += 3;
						break;
					case 0xD5:// PUSH DE
						mem_write(--sp, D_);
						mem_write(--sp, E_);
						pc++;
						break;
					case 0xD6:// SUB n
						{
							int value = mem_read(++pc);
							int result = A_ - value;
							int carrybits = A_ ^ value ^ result;
							if ((carrybits & 0x10) != 0 ) F_ |= 0x10;
							else F_ &=~ 0x10;
							if ((carrybits & 0x100) != 0) F_ |= 0x01;
							else F_ &=~ 0x01;
							if ((((carrybits << 1) ^ carrybits) & 0x100) != 0) F_ |= 0x04;
							else F_ &=~ 0x04;
							A_ = result&0xFF;
							if (A_ == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							if ((A_ & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							F_ |= 0x02;
							if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xD7:// RST 10H
						mem_write(--sp, (pc + 1) >> 8);
						mem_write(--sp, (pc + 1) & 0xFF);
						pc = 0x10;
						break;
					case 0xD8:// RET C
						if ((F_ & 0x01) == 0x01) {
							pc = (mem_read(sp + 1) << 8) + mem_read(sp);
							sp += 2;
							sp &= 0xFFFF;
							tstates -= 6;
						}
						else pc++;
						break;
					case 0xD9:// EXX
						{
							int temp;
							temp = B_;
							B_ = B2_;
							B2_ = temp;
							temp = C_;
							C_ = C2_;
							C2_ = temp;
							temp = D_;
							D_ = D2_;
							D2_ = temp;
							temp = E_;
							E_ = E2_;
							E2_ = temp;
							temp = H_;
							H_ = H2_;
							H2_ = temp;
							temp = L_;
							L_ = L2_;
							L2_ = temp;
						}
						pc++;
						break;
					case 0xDA:// JP C,(nn)
						if ((F_ & 0x01) == 0x01) pc = (mem_read(pc + 2) << 8) + mem_read(pc + 1);
						else pc+=3;
						break;
					case 0xDB:// IN A,(n)
						A_ = (in(mem_read(++pc))) & 0xFF;
						pc++;
						break;
					case 0xDC:// CALL C (nn)
						if ((F_ & 0x01) == 0x01) {
							mem_write(--sp, (pc + 3) >> 8);
							mem_write(--sp, (pc + 3) & 0xFF);
							pc = (mem_read(pc + 2) << 8) + mem_read(pc + 1);
							tstates -= 7;
						}
						else pc += 3;
						break;
					case 0xDD:// DD Opcode
						IX_ = doDDFD(mem_read(++pc), IX_);
						break;
					case 0xDE:// SBC A,n
						{
							int value = mem_read(++pc);
							int result = A_ - value - ((F_ & 0x01) == 0x01 ? 1 : 0);
							int carrybits = A_ ^ value ^ result;
							if ((carrybits & 0x10) != 0 ) F_ |= 0x10;
							else F_ &=~ 0x10;
							if ((carrybits & 0x100) != 0) F_ |= 0x01;
							else F_ &=~ 0x01;
							if ((((carrybits << 1) ^ carrybits) & 0x100) != 0) F_ |= 0x04;
							else F_ &=~ 0x04;
							A_ = result&0xFF;
							if (A_ == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							if ((A_ & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							F_ |= 0x02;
							if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xDF:// RST 18H
						mem_write(--sp, (pc + 1) >> 8);
						mem_write(--sp, (pc + 1) & 0xFF);
						pc = 0x18;
						break;
					case 0xE0:// RET PO
						if ((F_ & 0x04) != 0x04) {
							pc = (mem_read(sp + 1) << 8) + mem_read(sp);
							sp += 2;
							sp &= 0xFFFF;
							tstates -= 6;
						}
						else pc++;
						break;
					case 0xE1:// POP HL
						L_ = mem_read(sp);
						H_ = mem_read(sp + 1);
						sp += 2;
						sp &= 0xFFFF;
						pc++;
						break;
					case 0xE2:// JP PO,(nn)
						if ((F_ & 0x04) != 0x04) pc = (mem_read(pc + 2) << 8) + mem_read(pc + 1);
						else pc+=3;
						break;
					case 0xE3:// EX (SP),HL
						{
							int temp = ((H_ << 8) | L_);
							L_ = mem_read(sp);
							H_ = mem_read(sp + 1);
							mem_write(sp, temp & 0xFF);
							mem_write(sp+1, temp >> 8);
						}
						pc++;
						break;
					case 0xE4:// CALL PO (nn)
						if ((F_ & 0x04) != 0x04) {
							mem_write(--sp, (pc + 3) >> 8);
							mem_write(--sp, (pc + 3) & 0xFF);
							pc = (mem_read(pc + 2) << 8) + mem_read(pc + 1);
							tstates -= 7;
						}
						else pc += 3;
						break;
					case 0xE5:// PUSH HL
						mem_write(--sp, H_);
						mem_write(--sp, L_);
						pc++;
						break;
					case 0xE6:// AND (n)
						{
							int value = mem_read(++pc);
							A_ &= value;
							if ((A_ & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							if (A_ == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							F_ |= 0x10;
							if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
							F_ &=~ 0x02;
							F_ &=~ 0x01;
							if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xE7:// RST 20H
						mem_write(--sp, (pc + 1) >> 8);
						mem_write(--sp, (pc + 1) & 0xFF);
						pc=0x20;
						break;
					case 0xE8:// RET PE
						if ((F_ & 0x04) == 0x04) {
							pc = (mem_read(sp + 1) << 8) + mem_read(sp);
							sp += 2;
							sp &= 0xFFFF;
							tstates -= 6;
						}
						else pc++;
						break;
					case 0xE9:// JP (HL)
						pc = ((H_ << 8) | L_);
						break;
					case 0xEA:// JP PE,(nn)
						if ((F_ & 0x04) == 0x04) pc = (mem_read(pc + 2) << 8) + mem_read(pc + 1);
						else pc+=3;
						break;
					case 0xEB:// EX DE,HL
						{
							int temp = (D_ << 8) | E_;
							int tempp = (H_ << 8) | L_;
							E_ = tempp & 0xFF;
							D_ = (tempp >> 8) & 0xFF;
							L_ = temp & 0xFF;
							H_ = (temp >> 8) & 0xFF;
						}
						pc++;
						break;
					case 0xEC:// CALL PE (nn)
						if ((F_ & 0x04) == 0x04) {
							mem_write(--sp, (pc + 3) >> 8);
							mem_write(--sp, (pc + 3) & 0xFF);
							pc = (mem_read(pc + 2) << 8) + mem_read(pc + 1);
							tstates -= 7;
						}
						else pc += 3;
						break;
					case 0xED:// ED Opcode
						//doED(mem_read(++pc));
					{
						opcode = mem_read(++pc);
						boolean carry; // to preserve carry flag
						tstates -= opedstates[opcode];
						R_ = (R_&0x80) | ((R_+1)&0x7F);
						switch(opcode)
						{
							//  -- ED40 IN B,(C) -------------------------
							case 0x40:
								{
									int temp = in(C_);
									B_ = (temp) & 0xFF;
									if (temp == 0) F_ |= 0x40;
									else F_ &=~ 0x40;
									if ((temp & 0x80) != 0) F_ |= 0x80;
									else F_ &=~ 0x80;
									if(PARITY[temp & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;}; F_ &=~ 0x02; F_ &=~ 0x10;
								}
								pc++;
								break;
							//  -- ED41 OUT (C),B -------------------------
							case 0x41:
								out(C_, B_);
								pc++;
								break;
							// --  ED42 SBC HL, BC ------------------------ (C)
							case 0x42:
								{
									int value = ((B_ << 8) | C_);
									int reg = ((H_ << 8) | L_);
									int result = reg - value - (((F_ & 0x01) == 0x01) ? 1 : 0);
									int carrybits = reg ^ value ^ result;
									if ((carrybits & 0x1000) != 0 ) F_ |= 0x10;
									else F_ &=~ 0x10;
									if ((carrybits & 0x10000) != 0) F_ |= 0x01;
									else F_ &=~ 0x01;
									if ((((carrybits << 1) ^ carrybits) & 0x10000) != 0) F_ |= 0x04;
									else F_ &=~ 0x04;
									result &= 0xFFFF;
									if (result == 0) F_ |= 0x40;
									else F_ &=~ 0x40;
									if ((result & 0x8000) == 0) F_ &=~ 0x80;
									else F_ |= 0x80;
									F_ |= 0x02;
									L_   = result & 0xFF;
									H_  = (result >> 8) & 0xFF;
									result >>= 8;			// Bits 3 and 5 affected by H_ byte addition
									if((result & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;}
									if((result & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;}
								}
								pc++;
								break;
							//  -- ED43 LD (nn),BC ------------------------
							case 0x43:
								{
									int temp = (mem_read(pc + 2) << 8) + mem_read(pc + 1);
									mem_write(temp, C_);
									mem_write(temp + 1, B_);
								}
								pc+=3;
								break;
							//  -- ED44 NEG -------------------------------
							case 0x44:
							case 0x4C:
							case 0x54:
							case 0x5C:
							case 0x64:
							case 0x6C:
							case 0x74:
							case 0x7C:
								{
									int value = A_;
									A_ = 0;
									int result = A_ - value;
									int carrybits = A_ ^ value ^ result;
									if ((carrybits & 0x10) != 0 ) F_ |= 0x10;
									else F_ &=~ 0x10;
									if ((carrybits & 0x100) != 0) F_ |= 0x01;
									else F_ &=~ 0x01;
									if ((((carrybits << 1) ^ carrybits) & 0x100) != 0) F_ |= 0x04;
									else F_ &=~ 0x04;
									A_ = result&0xFF;
									if (A_ == 0) F_ |= 0x40;
									else F_ &=~ 0x40;
									if ((A_ & 0x80) == 0) F_ &=~ 0x80;
									else F_ |= 0x80;
									F_ |= 0x02;
									if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
									if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
								}
								pc++;
								break;
							//  -- ED45 RETN / RETI ------------------------------
							case 0x4D:
							case 0x5D:
							case 0x6D:
							case 0x7D:
								pc = (mem_read(sp + 1) << 8) + mem_read(sp);
								sp += 2;
								sp &= 0xFFFF;
								iff1 = iff2;
								break;
							case 0x45:
							case 0x55:
							case 0x65:
								pc = (mem_read(sp + 1) << 8) + mem_read(sp);
								sp += 2;
								sp &= 0xFFFF;
								iff1 = iff2;
								//checkInterrupt();
								break;
							//  -- ED46 IM 0-------------------------------
							case 0x46:
							case 0x4E:
							case 0x66:
							case 0x6E:
								//IM_ = 0;
								pc++;
								break;
							//  -- ED47 LD I, A ---------------------------
							case 0x47:
								I_ = A_;
								pc++;
								break;
							//  -- ED48 IN C,(C) -------------------------
							case 0x48:
								{
									int temp = in(C_);
									C_ = (temp) & 0xFF;
									if (temp == 0) F_ |= 0x40;
									else F_ &=~ 0x40;
									if ((temp & 0x80) != 0) F_ |= 0x80;
									else F_ &=~ 0x80;
									if(PARITY[temp & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;}; F_ &=~ 0x02; F_ &=~ 0x10;
								}
								pc++;
								break;
							//  -- ED49 OUT (C),C -------------------------
							case 0x49:
								out(C_, C_);
								pc++;
								break;
							//  -- ED4A ADC HL, BC ------------------------
							case 0x4A:
								{
									int value = ((B_ << 8) | C_);
									int reg = ((H_ << 8) | L_);
									int result = reg + value + (((F_ & 0x01) == 0x01) ? 1 : 0);
									int carrybits = reg ^ value ^ result;
									if ((carrybits & 0x1000) != 0 ) F_ |= 0x10;
									else F_ &=~ 0x10;
									if ((carrybits & 0x10000) != 0) F_ |= 0x01;
									else F_ &=~ 0x01;
									if ((((carrybits << 1) ^ carrybits) & 0x10000) != 0) F_ |= 0x04;
									else F_ &=~ 0x04;
									result &= 0xFFFF;
									if (result == 0) F_ |= 0x40;
									else F_ &=~ 0x40;
									if ((result & 0x8000) == 0) F_ &=~ 0x80;
									else F_ |= 0x80;
									F_ &=~ 0x02;
									L_   = result & 0xFF;
									H_  = (result >> 8) & 0xFF;
									result >>= 8;			// Bits 3 and 5 affected by H_ byte addition
									if((result & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;}
									if((result & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;}
								}
								pc++;
								break;
							//  -- ED4B LD BC, (nn) ----------------
							case 0x4B:
								{
									int temp = (mem_read(pc + 2) << 8) + mem_read(pc + 1);
									C_ = mem_read(temp);
									B_ = mem_read(temp + 1);
								}
								pc+=3;
								break;
							//  -- ED4F LD R, A ---------------------------
							case 0x4F:
								R_ = A_;
								pc++;
								break;
							//  -- ED50 IN D,(C) -------------------------
							case 0x50:
								{
									int temp = in(C_);
									D_ = (temp) & 0xFF;
									if (temp == 0) F_ |= 0x40;
									else F_ &=~ 0x40;
									if ((temp & 0x80) != 0) F_ |= 0x80;
									else F_ &=~ 0x80;
									if(PARITY[temp & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;}; F_ &=~ 0x02; F_ &=~ 0x10;
								}
								pc++;
								break;
							//  -- ED51 OUT (C),D -------------------------
							case 0x51:
								out(C_, D_);
								pc++;
								break;
							// --  ED52 SBC HL, DE ------------------------
							case 0x52:
								{
									int value = ((D_ << 8) | E_);
									int reg = ((H_ << 8) | L_);
									int result = reg - value - (((F_ & 0x01) == 0x01) ? 1 : 0);
									int carrybits = reg ^ value ^ result;
									if ((carrybits & 0x1000) != 0 ) F_ |= 0x10;
									else F_ &=~ 0x10;
									if ((carrybits & 0x10000) != 0) F_ |= 0x01;
									else F_ &=~ 0x01;
									if ((((carrybits << 1) ^ carrybits) & 0x10000) != 0) F_ |= 0x04;
									else F_ &=~ 0x04;
									result &= 0xFFFF;
									if (result == 0) F_ |= 0x40;
									else F_ &=~ 0x40;
									if ((result & 0x8000) == 0) F_ &=~ 0x80;
									else F_ |= 0x80;
									F_ |= 0x02;
									L_   = result & 0xFF;
									H_  = (result >> 8) & 0xFF;
									result >>= 8;			// Bits 3 and 5 affected by H_ byte addition
									if((result & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;}
									if((result & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;}
								}
								pc++;
								break;
							//  -- ED53 LD (nn),DE ------------------------
							case 0x53:
								{
									int temp = (mem_read(pc + 2) << 8) + mem_read(pc + 1);
									mem_write(temp, E_);	 //SPl
									mem_write(temp + 1, D_);   //SPh
								}
								pc+=3;
								break;
							//  -- ED56 IM 1-------------------------------
							case 0x56:
								//IM_ = 1;
								pc++;
								break;
							//  -- ED57 LD A, I ---------------------------
							case 0x57:
								A_ = I_ & 0xFF;
								if (iff2) F_ |= 0x04;
								else F_ &=~ 0x04;
								if (I_ == 0) F_ |= 0x40;
								else F_ &=~ 0x40;
								if ((I_ & 0x80) != 0) F_ |= 0x80;
								else F_ &=~ 0x80;
								F_ &=~ 0x10; F_ &=~ 0x02;
								pc++;
								break;
							//  -- ED58 IN E,(C) -------------------------
							case 0x58:
								{
									int temp = in(C_);
									E_ = (temp) & 0xFF;
									if (temp == 0) F_ |= 0x40;
									else F_ &=~ 0x40;
									if ((temp & 0x80) != 0) F_ |= 0x80;
									else F_ &=~ 0x80;
									if(PARITY[temp & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;}; F_ &=~ 0x02; F_ &=~ 0x10;
								}
								pc++;
								break;
							//  -- ED59 OUT (C),E -------------------------
							case 0x59:
								out(C_, E_);
								pc++;
								break;
							//  -- ED5A ADC HL, DE ------------------------
							case 0x5A:
								{
									int value = ((D_ << 8) | E_);
									int reg = ((H_ << 8) | L_);
									int result = reg + value + (((F_ & 0x01) == 0x01) ? 1 : 0);
									int carrybits = reg ^ value ^ result;
									if ((carrybits & 0x1000) != 0 ) F_ |= 0x10;
									else F_ &=~ 0x10;
									if ((carrybits & 0x10000) != 0) F_ |= 0x01;
									else F_ &=~ 0x01;
									if ((((carrybits << 1) ^ carrybits) & 0x10000) != 0) F_ |= 0x04;
									else F_ &=~ 0x04;
									result &= 0xFFFF;
									if (result == 0) F_ |= 0x40;
									else F_ &=~ 0x40;
									if ((result & 0x8000) == 0) F_ &=~ 0x80;
									else F_ |= 0x80;
									F_ &=~ 0x02;
									L_   = result & 0xFF;
									H_  = (result >> 8) & 0xFF;
									result >>= 8;			// Bits 3 and 5 affected by H_ byte addition
									if((result & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;}
									if((result & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;}
								}
								pc++;
								break;
							//  -- ED5B LD DE, (nn) -----------------------
							case 0x5B:
								{
									int temp = (mem_read(pc + 2) << 8) + mem_read(pc + 1);
									E_ = mem_read(temp);
									D_ = mem_read(temp + 1);
								}
								pc+=3;
								break;
							//  -- ED5E IM 2-------------------------------
							case 0x5E:
								//IM_ = 2;
								pc++;
								break;
							// -- ED5F LD A,R -----------------------------
							case 0x5F:
								{
									int temp = R_;
									A_ = temp & 0xFF;
									if (iff2) F_ |= 0x04;
									else F_ &=~ 0x04;
									if (temp == 0) F_ |= 0x40;
									else F_ &=~ 0x40;
									if ((temp & 0x80) != 0) F_ |= 0x80;
									else F_ &=~ 0x80;
									F_ &=~ 0x10; F_ &=~ 0x02;
								}
								pc++;
								break;
							//  -- ED60 IN H,(C) -------------------------
							case 0x60:
								{
									int temp = in(C_);
									H_ = (temp) & 0xFF;
									if (temp == 0) F_ |= 0x40;
									else F_ &=~ 0x40;
									if ((temp & 0x80) == 0x80) F_ |= 0x80;
									else F_ &=~ 0x80;
									if(PARITY[temp & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;}; F_ &=~ 0x02; F_ &=~ 0x10;
								}
								pc++;
								break;
							//  -- ED61 OUT (C),H -------------------------
							case 0x61:
								out(C_, H_);
								pc++;
								break;
							// --  ED62 SBC HL, HL ------------------------
							case 0x62:
								{
									int value = ((H_ << 8) | L_);
									int reg = ((H_ << 8) | L_);
									int result = reg - value - (((F_ & 0x01) == 0x01) ? 1 : 0);
									int carrybits = reg ^ value ^ result;
									if ((carrybits & 0x1000) != 0 ) F_ |= 0x10;
									else F_ &=~ 0x10;
									if ((carrybits & 0x10000) != 0) F_ |= 0x01;
									else F_ &=~ 0x01;
									if ((((carrybits << 1) ^ carrybits) & 0x10000) != 0) F_ |= 0x04;
									else F_ &=~ 0x04;
									result &= 0xFFFF;
									if (result == 0) F_ |= 0x40;
									else F_ &=~ 0x40;
									if ((result & 0x8000) == 0) F_ &=~ 0x80;
									else F_ |= 0x80;
									F_ |= 0x02;
									L_   = result & 0xFF;
									H_  = (result >> 8) & 0xFF;
									result >>= 8;			// Bits 3 and 5 affected by H_ byte addition
									if((result & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;}
									if((result & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;}
								}
								pc++;
								break;
							//  -- ED63 LD (nn),HL ------------------------
							case 0x63:
								mem_write((mem_read(pc + 2) << 8) + mem_read(pc + 1), L_);	   //SPl
								mem_write(((mem_read(pc + 2) << 8) + mem_read(pc + 1) + 1), H_);   //SPh
								pc+=3;
								break;
							//  -- ED67 RRD -------------------------------
							case 0x67:
								{
									int temp = A_;
									int tempp = mem_read(((H_ << 8) | L_));
									// move high 4 of hl to low 4 of hl
									// move low 4 of a to high 4 of hl
									mem_write(((H_ << 8) | L_), (tempp >> 4) | ((temp & 0x0F) << 4));
									// move 4 lowest bits of hl to low 4 of a
									temp = (temp & 0xF0) | (tempp & 0x0F);
									A_ = temp & 0xFF;
									if (temp == 0) F_ |= 0x40;
									else F_ &=~ 0x40;
									if ((temp & 0x80) == 0) F_ &=~ 0x80;
									else F_ |= 0x80;
									if(PARITY[temp & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
									F_ &=~ 0x02; F_ &=~ 0x10;
								}
								pc++;
								break;
							//  -- ED68 IN L,(C) --------------------------
							case 0x68:
								{
									int temp = in(C_);
									L_ = (temp) & 0xFF;
									if (temp == 0) F_ |= 0x40;
									else F_ &=~ 0x40;
									if ((temp & 0x80) == 0x80) F_ |= 0x80;
									else F_ &=~ 0x80;
									if(PARITY[temp & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;}; F_ &=~ 0x02; F_ &=~ 0x10;
								}
								pc++;
								break;
							//  -- ED69 OUT (C),L -------------------------
							case 0x69:
								out(C_, L_);
								pc++;
								break;
							//  -- ED6A ADC HL, HL ------------------------
							case 0x6A:
								{
									int value = ((H_ << 8) | L_);
									int reg = ((H_ << 8) | L_);
									int result = reg + value + (((F_ & 0x01) == 0x01) ? 1 : 0);
									int carrybits = reg ^ value ^ result;
									if ((carrybits & 0x1000) != 0 ) F_ |= 0x10;
									else F_ &=~ 0x10;
									if ((carrybits & 0x10000) != 0) F_ |= 0x01;
									else F_ &=~ 0x01;
									if ((((carrybits << 1) ^ carrybits) & 0x10000) != 0) F_ |= 0x04;
									else F_ &=~ 0x04;
									result &= 0xFFFF;
									if (result == 0) F_ |= 0x40;
									else F_ &=~ 0x40;
									if ((result & 0x8000) == 0) F_ &=~ 0x80;
									else F_ |= 0x80;
									F_ &=~ 0x02;
									L_   = result & 0xFF;
									H_  = (result >> 8) & 0xFF;
									result >>= 8;			// Bits 3 and 5 affected by H_ byte addition
									if((result & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;}
									if((result & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;}
								}
								pc++;
								break;
							//  -- ED6B LD HL,(nn) -----------------------
							case 0x6B:
								{
									int temp = (mem_read(pc + 2) << 8) + mem_read(pc + 1);
									L_ = mem_read(temp);
									H_ = mem_read(temp + 1);
								}
								pc+=3;
								break;
							//  -- ED6F RLD -------------------------------
							case 0x6F:
								{
									int temp = A_;
									int tempp = mem_read(((H_ << 8) | L_));
									// move low 4 of hl to high 4 of hl
									// move low 4 of a to low 4 of hl
									mem_write(((H_ << 8) | L_), (tempp&0x0F)<<4 | (temp&0x0F));
									// move high 4 of hl to low 4 of a
									temp = (temp&0xF0) | (tempp>>4);
									A_ = temp & 0xFF;
									if (temp == 0) F_ |= 0x40;
									else F_ &=~ 0x40;
									if ((temp & 0x80) == 0) F_ &=~ 0x80;
									else F_ |= 0x80;
									F_ &=~ 0x02; F_ &=~ 0x10;
									if(PARITY[temp & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
								}
								pc++;
								break;
							//  *- ED71 OUT (C),0 -------------------------
							case 0x71:
								out(C_, 0);
								pc++;
								break;
							// --  ED72 SBC HL, SP ------------------------
							case 0x72:
								{
									int value = sp;
									int reg = ((H_ << 8) | L_);
									int result = reg - value - (((F_ & 0x01) == 0x01) ? 1 : 0);
									int carrybits = reg ^ value ^ result;
									if ((carrybits & 0x1000) != 0 ) F_ |= 0x10;
									else F_ &=~ 0x10;
									if ((carrybits & 0x10000) != 0) F_ |= 0x01;
									else F_ &=~ 0x01;
									if ((((carrybits << 1) ^ carrybits) & 0x10000) != 0) F_ |= 0x04;
									else F_ &=~ 0x04;
									result &= 0xFFFF;
									if (result == 0) F_ |= 0x40;
									else F_ &=~ 0x40;
									if ((result & 0x8000) == 0) F_ &=~ 0x80;
									else F_ |= 0x80;
									F_ |= 0x02;
									L_   = result & 0xFF;
									H_  = (result >> 8) & 0xFF;
									result >>= 8;			// Bits 3 and 5 affected by H_ byte addition
									if((result & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;}
									if((result & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;}
								}
								pc++;
								break;
							//  -- ED73 LD (nn),SP ------------------------
							case 0x73:
								mem_write((mem_read(pc + 2) << 8) + mem_read(pc + 1), sp & 0xFF);	 //SPl
								mem_write(((mem_read(pc + 2) << 8) + mem_read(pc + 1) + 1), sp >> 8);   //SPh
								pc+=3;
								break;
							//  -- ED78 IN A,(C) -------------------------
							// Output reg A to Port stored in reg c
							case 0x78:
								{
									int temp = in(C_);
									A_ = temp & 0xFF;
									if (temp == 0) F_ |= 0x40;
									else F_ &=~ 0x40;
									if ((temp & 0x80) == 0x80) F_ |= 0x80;
									else F_ &=~ 0x80;
									if(PARITY[temp & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
									F_ &=~ 0x02;
									F_ &=~ 0x10;
								}
								pc++;
								break;
							//  -- ED79 OUT (C),A -------------------------
							// Output reg A to Port stored in reg c
							case 0x79:
								out(C_, A_);
								pc++;
								break;
							// --  ED7A ADC HL, SP ------------------------
							case 0x7A:
								{
									int value = sp;
									int reg = ((H_ << 8) | L_);
									int result = reg + value + (((F_ & 0x01) == 0x01) ? 1 : 0);
									int carrybits = reg ^ value ^ result;
									if ((carrybits & 0x1000) != 0 ) F_ |= 0x10;
									else F_ &=~ 0x10;
									if ((carrybits & 0x10000) != 0) F_ |= 0x01;
									else F_ &=~ 0x01;
									if ((((carrybits << 1) ^ carrybits) & 0x10000) != 0) F_ |= 0x04;
									else F_ &=~ 0x04;
									result &= 0xFFFF;
									if (result == 0) F_ |= 0x40;
									else F_ &=~ 0x40;
									if ((result & 0x8000) == 0) F_ &=~ 0x80;
									else F_ |= 0x80;
									F_ &=~ 0x02;
									L_   = result & 0xFF;
									H_  = (result >> 8) & 0xFF;
									result >>= 8;			// Bits 3 and 5 affected by H_ byte addition
									if((result & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;}
									if((result & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;}
								}
								pc++;
								break;
							//  -- ED7B LD SP, (nn) -----------------------
							case 0x7B:
								{
									int temp = mem_read((mem_read(pc + 2) << 8) + mem_read(pc + 1)); // SPl
									int tempp = mem_read((mem_read(pc + 2) << 8) + mem_read(pc + 1) + 1); // SPh
									sp = temp | (tempp << 8);
								}
								pc+=3;
								break;
							//  -- EDA0 LDI ---------------------------------- (c)
							case 0xA0:
								// (DE) <- (HL)
								mem_write(((D_ << 8) | E_), mem_read(((H_ << 8) | L_)));
								E_ = (E_ + 1) & 0xFF;
								if (E_ == 0)
									D_ = (D_ + 1) & 0xFF;
									L_ = (L_ + 1) & 0xFF;
								if (L_ == 0)
									H_ = (H_ + 1) & 0xFF;
									C_ = (C_ - 1) & 0xFF;
								if (C_ == 255) B_ = (B_ - 1) & 0xFF;
								if (((B_ << 8) | C_) == 0) F_ &=~ 0x04;
								else F_ |= 0x04;
								F_ &=~ 0x02; F_ &=~ 0x10;
								pc++;
								break;
							//  -- EDA1 CPI ------------------------------
							case 0xA1:
								carry = ((F_ & 0x01) == 0x01);
								{
									int value = mem_read((H_ << 8) | L_);
									int result = A_ - value;
									int carrybits = A_ ^ value ^ result;
									if ((carrybits & 0x10) != 0 ) F_ |= 0x10;
									else F_ &=~ 0x10;
									if ((carrybits & 0x100) != 0) F_ |= 0x01;
									else F_ &=~ 0x01;
									if ((((carrybits << 1) ^ carrybits) & 0x100) != 0) F_ |= 0x04;
									else F_ &=~ 0x04;
									result &= 0xFF;
									if ((result & 0x80) == 0) F_ &=~ 0x80;
									else F_ |= 0x80;
									if (result == 0) F_ |= 0x40;
									else F_ &=~ 0x40;
									F_ |= 0x02;
									if((value & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
									if((value & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
								}
								L_ = (L_ + 1) & 0xFF;if (L_ == 0) H_ = (H_ + 1) & 0xFF;
								C_ = (C_ - 1) & 0xFF;if (C_ == 255) B_ = (B_ - 1) & 0xFF;
								if(((B_ << 8) | C_) == 0) F_ &=~ 0x04;
								else F_ |= 0x04;
								F_ |= 0x02;
								if (carry) F_ |= 0x01;
								else F_ &=~ 0x01;
								pc++;
								break;
							//  -- EDA2 INI -------------------------------
							case 0xA2:
								{
									int temp = in(C_);
									mem_write(((H_ << 8) | L_), temp);
									B_ = dec8(B_);
									L_ = (L_ + 1) & 0xFF;if (L_ == 0) H_ = (H_ + 1) & 0xFF;
									if (temp + ((B_+1)&0xFF) > 255) {
										F_ |= 0x01;
										F_ |= 0x10;
									} else {
										F_ &=~ 0x01;
										F_ &=~ 0x10;
									}
									if ((temp & 0x80) == 0x80) F_ |= 0x02;
									else F_ &=~ 0x02;
								}
								pc++;
								break;
							//  -- EDA3 OUTI ------------------------------
							// see p14 of undocumented z80 for additional flag info
							case 0xA3:
								{
									int temp = mem_read((H_ << 8) | L_);
									out(C_, temp);
									L_ = (L_ + 1) & 0xFF;if (L_ == 0) H_ = (H_ + 1) & 0xFF;
									B_ = dec8(B_); // Flags in OUTI adjusted in same way as dec b anyway.
									if ((L_ + temp) > 255) {
										F_ |= 0x01;
										F_ |= 0x10;
									} else {
										F_ &=~ 0x01;
										F_ &=~ 0x10;
									}
									if ((temp & 0x80) == 0x80) F_ |= 0x02;
									else F_ &=~ 0x02;
								}
								pc++;
								break;
							//  -- EDA8 LDD ----------------------------------
							case 0xA8:
								mem_write(((D_ << 8) | E_), mem_read((H_ << 8) | L_));
								E_ = (E_ - 1) & 0xFF;
								if (E_ == 255)
									D_ = (D_ - 1) & 0xFF;
									L_ = (L_ - 1) & 0xFF;
								if (L_ == 255)
									H_ = (H_ - 1) & 0xFF;
									C_ = (C_ - 1) & 0xFF;
								if (C_ == 255) B_ = (B_ - 1) & 0xFF;
								if (((B_ << 8) | C_) == 0) F_ &=~ 0x04;
								else F_ |= 0x04;
								F_ &=~ 0x02; F_ &=~ 0x10;
								pc++;
								break;
							//  -- EDA9 CPD ------------------------------
							case 0xA9:
								carry = ((F_ & 0x01) == 0x01);
								{
									int value = mem_read(((H_ << 8) | L_));
									int result = A_ - value;
									int carrybits = A_ ^ value ^ result;
									if ((carrybits & 0x10) != 0 ) F_ |= 0x10;
									else F_ &=~ 0x10;
									if ((carrybits & 0x100) != 0) F_ |= 0x01;
									else F_ &=~ 0x01;
									if ((((carrybits << 1) ^ carrybits) & 0x100) != 0) F_ |= 0x04;
									else F_ &=~ 0x04;
									result &= 0xFF;
									if ((result & 0x80) == 0) F_ &=~ 0x80;
									else F_ |= 0x80;
									if (result == 0) F_ |= 0x40;
									else F_ &=~ 0x40;
									F_ |= 0x02;
									if((value & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
									if((value & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
								}
								L_ = (L_ - 1) & 0xFF;if (L_ == 255) H_ = (H_ - 1) & 0xFF;
								C_ = (C_ - 1) & 0xFF;if (C_ == 255) B_ = (B_ - 1) & 0xFF;
								if(((B_ << 8) | C_) == 0) F_ &=~ 0x04;
								else F_ |= 0x04;
								F_ |= 0x02;
								if (carry) F_ |= 0x01;
								else F_ &=~ 0x01;
								pc++;
								break;
							//  -- EDAA IND -------------------------------
							case 0xAA:
								{
									int temp = in(C_);
									mem_write(((H_ << 8) | L_), temp);
									B_ = dec8(B_);
									L_ = (L_ - 1) & 0xFF;if (L_ == 255) H_ = (H_ - 1) & 0xFF;
									if (temp + ((B_-1)&0xFF) > 255) {
										F_ |= 0x01;
										F_ |= 0x10;
									} else {
										F_ &=~ 0x01;
										F_ &=~ 0x10;
									}
									if ((temp & 0x80) == 0x80) F_ |= 0x02;
									else F_ &=~ 0x02;
								}
								pc++;
								break;
							//  -- EDAB OUTD ------------------------------
							// see p14 of undocumented z80 for additional flag info
							case 0xAB:
								{
									int temp = mem_read(((H_ << 8) | L_));
									out(C_, temp);
									L_ = (L_ - 1) & 0xFF;if (L_ == 255) H_ = (H_ - 1) & 0xFF;
									B_ = dec8(B_); // Flags in OUTI adjusted in same way as dec b anyway.
									if ((L_ + temp) > 255) {
										F_ |= 0x01;
										F_ |= 0x10;
									} else {
										F_ &=~ 0x01;
										F_ &=~ 0x10;
									}
									if ((temp & 0x80) == 0x80) F_ |= 0x02;
									else F_ &=~ 0x02;
								}
								pc++;
								break;
							//  -- EDB0 LDIR ------------------------------
							case 0xB0:
								mem_write(((D_ << 8) | E_), mem_read(((H_ << 8) | L_)));
								E_ = (E_ + 1) & 0xFF;if (E_ == 0) D_ = (D_ + 1) & 0xFF;
								L_ = (L_ + 1) & 0xFF;if (L_ == 0) H_ = (H_ + 1) & 0xFF;
								C_ = (C_ - 1) & 0xFF;if (C_ == 255) B_ = (B_ - 1) & 0xFF;
								if (((B_ << 8) | C_) != 0) {
									F_ |= 0x04;
									tstates -= 5;
									pc--;
								} else {
									F_ &=~ 0x04;
									pc++;
								}
								F_ &=~ 0x02; F_ &=~ 0x10;
								break;
							//  -- EDB1 CPIR ------------------------------
							case 0xB1:
								carry = ((F_ & 0x01) == 0x01);
								{
									int value = mem_read(((H_ << 8) | L_));
									int result = A_ - value;
									int carrybits = A_ ^ value ^ result;
									if ((carrybits & 0x10) != 0 ) F_ |= 0x10;
									else F_ &=~ 0x10;
									if ((carrybits & 0x100) != 0) F_ |= 0x01;
									else F_ &=~ 0x01;
									if ((((carrybits << 1) ^ carrybits) & 0x100) != 0) F_ |= 0x04;
									else F_ &=~ 0x04;
									result &= 0xFF;
									if ((result & 0x80) == 0) F_ &=~ 0x80;
									else F_ |= 0x80;
									if (result == 0) F_ |= 0x40;
									else F_ &=~ 0x40;
									F_ |= 0x02;
									if((value & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
									if((value & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
								}
								L_ = (L_ + 1) & 0xFF;if (L_ == 0) H_ = (H_ + 1) & 0xFF;
								C_ = (C_ - 1) & 0xFF;if (C_ == 255) B_ = (B_ - 1) & 0xFF;
								if (((B_ << 8) | C_) == 0) F_ &=~ 0x04;
								else F_ |= 0x04;
								if ((((F_ & 0x04) == 0x04)) && (!((F_ & 0x40) == 0x40))) {
									tstates -=5;
									pc--;
								} else {
									pc++;
								}
								if (carry) F_ |= 0x01;
								else F_ &=~ 0x01;
								F_ |= 0x02; // Sign set by the cp instruction
								break;
							//  -- EDB2 INIR ------------------------------
							case 0xB2:
								{
									int temp = in(C_);
									mem_write(((H_ << 8) | L_), temp);
									B_ = dec8(B_);
									L_ = (L_ + 1) & 0xFF;if (L_ == 0) H_ = (H_ + 1) & 0xFF;
									if (B_ != 0) {
										tstates -= 5;
										pc--;
									} else {
										pc++;
									}
									if (temp + ((B_+1)&0xFF) > 255) {
										F_ |= 0x01;
										F_ |= 0x10;
									} else {
										F_ &=~ 0x01;
										F_ &=~ 0x10;
									}
									if ((temp & 0x80) == 0x80) F_ |= 0x02;
									else F_ &=~ 0x02;
								}
								break;
							//  -- EDB3 OTIR ------------------------------
							case 0xB3:
								{
									int temp = mem_read(((H_ << 8) | L_));
									out(C_, temp);
									B_ = dec8(B_);
									L_ = (L_ + 1) & 0xFF;if (L_ == 0) H_ = (H_ + 1) & 0xFF;
									if (B_ != 0) {
										tstates -= 5;
										pc--;
									} else {
										pc++;
									}
									if ((L_ + temp) > 255) {
										F_ |= 0x01;
										F_ |= 0x10;
									} else {
										F_ &=~ 0x01;
										F_ &=~ 0x10;
									}
									if ((temp & 0x80) == 0x80) F_ |= 0x02;
									else F_ &=~ 0x02;
								}
								break;

							// -- EDB8 LDDR ---------------------------------
							case 0xB8:
								mem_write(((D_ << 8) | E_), mem_read(((H_ << 8) | L_)));
								E_ = (E_ - 1) & 0xFF;
								if (E_ == 255)
									D_ = (D_ - 1) & 0xFF;
									L_ = (L_ - 1) & 0xFF;
								if (L_ == 255)
									H_ = (H_ - 1) & 0xFF;
									C_ = (C_ - 1) & 0xFF;
								if (C_ == 255) B_ = (B_ - 1) & 0xFF;
								if (((B_ << 8) | C_) != 0) {
									F_ |= 0x04;
									tstates -= 5;
									pc--;
								} else {
									F_ &=~ 0x04;
									pc++;
								}
								F_ &=~ 0x02; F_ &=~ 0x10;
								break;
							// -- EDB9 CPDR ------------------------------------
							case 0xB9:
								carry = ((F_ & 0x01) == 0x01);
								{
									int value = mem_read(((H_ << 8) | L_));
									int result = A_ - value;
									int carrybits = A_ ^ value ^ result;
									if ((carrybits & 0x10) != 0 ) F_ |= 0x10;
									else F_ &=~ 0x10;
									if ((carrybits & 0x100) != 0) F_ |= 0x01;
									else F_ &=~ 0x01;
									if ((((carrybits << 1) ^ carrybits) & 0x100) != 0) F_ |= 0x04;
									else F_ &=~ 0x04;
									result &= 0xFF;
									if ((result & 0x80) == 0) F_ &=~ 0x80;
									else F_ |= 0x80;
									if (result == 0) F_ |= 0x40;
									else F_ &=~ 0x40;
									F_ |= 0x02;
									if((value & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
									if((value & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
								}
								L_ = (L_ - 1) & 0xFF;if (L_ == 255) H_ = (H_ - 1) & 0xFF;
								C_ = (C_ - 1) & 0xFF;if (C_ == 255) B_ = (B_ - 1) & 0xFF;
								if (((B_ << 8) | C_) == 0) F_ &=~ 0x04;
								else F_ |= 0x04;
								if ((((F_ & 0x04) == 0x04)) && (!((F_ & 0x40) == 0x40))) {
									tstates -=5;
									pc--;
								} else {
									pc++;
								}
								if (carry) F_ |= 0x01;
								else F_ &=~ 0x01;
								F_ |= 0x02;
								break;
							//  -- EDBA INDR ------------------------------
							case 0xBA:
								{
									int temp = in(C_);
									mem_write(((H_ << 8) | L_), temp);
									B_ = dec8(B_);
									L_ = (L_ - 1) & 0xFF;if (L_ == 255) H_ = (H_ - 1) & 0xFF;
									if (B_ != 0) {
										tstates -= 5;
										pc--;
									} else {
										pc++;
									}
									if (temp + ((B_-1)&0xFF) > 255) {
										F_ |= 0x01;
										F_ |= 0x10;
									} else {
										F_ &=~ 0x01;
										F_ &=~ 0x10;
									}
									if ((temp & 0x80) == 0x80) F_ |= 0x02;
									else F_ &=~ 0x02;
								}
								break;
							//  -- EDBB OTDR ------------------------------
							case 0xBB:
								{
									int temp = mem_read(((H_ << 8) | L_));
									out(C_, temp);
									B_ = dec8(B_);
									L_ = (L_ - 1) & 0xFF;if (L_ == 255) H_ = (H_ - 1) & 0xFF;
									if (B_ != 0) {
										tstates -= 5;
										pc--;
									} else {
										pc++;
									}
									if ((L_ + temp) > 255) {
										F_ |= 0x01;
										F_ |= 0x10;
									} else {
										F_ &=~ 0x01;
										F_ &=~ 0x10;
									}
									if ((temp & 0x80) == 0x80) F_ |= 0x02;
									else F_ &=~ 0x02;
								}
								break;
								// -- Unimplented ED Opcode --------------------
							default:
								System.out.println("Unimplemented ED Opcode : 0x"+Integer.toHexString(opcode).toUpperCase());
								//pc++;
								break;
						} // end of switch
					} // end of ed ops
						break;
					case 0xEE:// XOR n
						{
							int value = mem_read(++pc);
							A_ ^= value; // bitwise exclusive OR
							if (A_ == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							if ((A_ & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							F_ &=~ 0x10;
							F_ &=~ 0x02;
							F_ &=~ 0x01;
							if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
							if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xEF:// RST 28H
						mem_write(--sp, (pc + 1) >> 8);
						mem_write(--sp, (pc + 1) & 0xFF);
						pc = 0x28;
						break;
					case 0xF0:// RET P
						if ((F_ & 0x80) != 0x80) {
							pc = (mem_read(sp + 1) << 8) + mem_read(sp);
							sp += 2;
							sp &= 0xFFFF;
							tstates -= 6;
						}
						else pc++;
						break;
					case 0xF1:// POP AF
						F_ = mem_read(sp);
						A_ = mem_read(sp+1);
						sp+=2;
						pc++;
						break;
					case 0xF2:// JP P,(nn)
						if ((F_ & 0x80) != 0x80) pc = (mem_read(pc + 2) << 8) + mem_read(pc + 1);
						else pc+=3;
						break;
					case 0xF3:// DI
						iff1 = false;
						iff2 = false;
						EI_inst = true;
						pc++;
						break;
					case 0xF4:// CALL P (nn)
						if ((F_ & 0x80) != 0x80) {
							mem_write(--sp, (pc + 3) >> 8);
							mem_write(--sp, (pc + 3) & 0xFF);
							pc = (mem_read(pc + 2) << 8) + mem_read(pc + 1);
							tstates -= 7;
						}
						else pc += 3;
						break;
					case 0xF5:// PUSH AF
						mem_write(--sp, ((A_ << 8) + F_) >> 8);
						mem_write(--sp, ((A_ << 8) + F_) & 0xFF);
						pc++;
						break;
					case 0xF6:// OR n
						{
							int value = mem_read(++pc);
							A_ |= value; // bitwise inclusive OR to turn relevant bits on
							if ((A_ & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							if (A_ == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							F_ &=~ 0x10;
							if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
							F_ &=~ 0x02;
							F_ &=~ 0x01;
							if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xF7:// RST 30H
						mem_write(--sp, (pc + 1) >> 8);
						mem_write(--sp, (pc + 1) & 0xFF);
						pc=0x30;
						break;
					case 0xF8:// RET M
						if ((F_ & 0x80) == 0x80) {
							pc = (mem_read(sp + 1) << 8) + mem_read(sp);
							sp += 2;
							sp &= 0xFFFF;
							tstates -= 6;
						}
						else pc++;
						break;
					case 0xF9:// LD SP,HL
						sp = ((H_ << 8) | L_);
						pc++;
						break;
					case 0xFA:// JP M,(nn)
						if ((F_ & 0x80) == 0x80) pc = (mem_read(pc + 2) << 8) + mem_read(pc + 1);
						else pc+=3;
						break;
					case 0xFB:// EI
						iff1 = true;
						iff2 = true;
						EI_inst = true;
						//checkInterrupt();
						pc++;
						break;
					case 0xFC:// CALL M (nn)
						if ((F_ & 0x80) == 0x80) {
							mem_write(--sp, (pc + 3) >> 8);
							mem_write(--sp, (pc + 3) & 0xFF);
							pc = (mem_read(pc + 2) << 8) + mem_read(pc + 1);
							tstates -= 7;
						}
						else pc += 3;
						break;
					case 0xFD:// FD Opcode
						IY_ = doDDFD(mem_read(++pc), IY_);
						break;
					case 0xFE:// CP n
						{
							int value = mem_read(++pc);
							int result = A_ - value;
							int carrybits = A_ ^ value ^ result;
							if ((carrybits & 0x10) != 0 ) F_ |= 0x10;
							else F_ &=~ 0x10;
							if ((carrybits & 0x100) != 0) F_ |= 0x01;
							else F_ &=~ 0x01;
							if ((((carrybits << 1) ^ carrybits) & 0x100) != 0) F_ |= 0x04;
							else F_ &=~ 0x04;
							result &= 0xFF;
							if ((result & 0x80) == 0) F_ &=~ 0x80;
							else F_ |= 0x80;
							if (result == 0) F_ |= 0x40;
							else F_ &=~ 0x40;
							F_ |= 0x02;
							if((value & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
							if((value & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
						}
						pc++;
						break;
					case 0xFF:// RST 38H
						mem_write(--sp, (pc + 1) >> 8);
						mem_write(--sp, (pc + 1) & 0xFF);
						pc=0x38;
						break;
					// Unimplemented Opcode
					default:
						System.out.println("Unimplemented Opcode : 0x"+Integer.toHexString(opcode&0xff).toUpperCase());
						pc++;
						break;
				} // end switch
			} // end doOp
		} // end while
	} // end Method Z80



	private final int dec8(int value){
		value = (value - 1) & 0xFF;
		if ((value & 0x0F) == 0x0F) F_ |= 0x10;
		else F_ &=~ 0x10;
		if (value == 0x7F) F_ |= 0x04;
		else F_ &=~ 0x04;
		F_ |= 0x02;
		if (value == 0) F_ |= 0x40;
		else F_ &=~ 0x40;
		if ((value & 0x80) == 0) F_ &=~ 0x80;
		else F_ |= 0x80;
		return value;
	}



	private final int inc8(int value) {
		value = (value + 1) & 0xFF;
		if ((value & 0x0F) == 0x00) F_ |= 0x10;
		else F_ &=~ 0x10;
		if (value == 0x80) F_ |= 0x04;
		else F_ &=~ 0x04;
		F_ &=~ 0x02;
		if (value == 0) F_ |= 0x40;
		else F_ &=~ 0x40;
		if ((value & 0x80) == 0) F_ &=~ 0x80;
		else F_ |= 0x80;
		return value;
	}



	private final int add(int reg, int value) {
		return flag_a(reg, value, reg + value);
	}

	private final int adc(int reg, int value) {
		return flag_a(reg, value, reg + value + ((F_ & 0x01) == 0x01 ? 1 : 0));
	}

	private final int sub(int reg, int value) {
		return flag_s(reg, value, reg - value);
	}

	private final int sbc(int reg, int value) {
		return flag_s(reg, value, reg - value - ((F_ & 0x01) == 0x01 ? 1 : 0));
	}

	private final int flag_a(int reg, int value, int result) {
		int carrybits = reg ^ value ^ result;
		if ((carrybits & 0x10) != 0 ) F_ |= 0x10;
		else F_ &=~ 0x10;
		if ((carrybits & 0x100) != 0) F_ |= 0x01;
		else F_ &=~ 0x01;
		if ((((carrybits << 1) ^ carrybits) & 0x100) != 0) F_ |= 0x04;
		else F_ &=~ 0x04;
		reg = result & 0xFF;
		if (reg == 0) F_ |= 0x40;
		else F_ &=~ 0x40;
		if ((reg & 0x80) == 0) F_ &=~ 0x80;
		else F_ |= 0x80;
		F_ &=~ 0x02;
		if((reg & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
		if((reg & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
		return reg;
	}

	private final int flag_s(int reg, int value, int result) {
		int carrybits = reg ^ value ^ result;
		if ((carrybits & 0x10) != 0 ) F_ |= 0x10;
		else F_ &=~ 0x10;
		if ((carrybits & 0x100) != 0) F_ |= 0x01;
		else F_ &=~ 0x01;
		if ((((carrybits << 1) ^ carrybits) & 0x100) != 0) F_ |= 0x04;
		else F_ &=~ 0x04;
		reg = result & 0xFF;
		if (reg == 0) F_ |= 0x40;
		else F_ &=~ 0x40;
		if ((reg & 0x80) == 0) F_ &=~ 0x80;
		else F_ |= 0x80;
		F_ |= 0x02;
		if((reg & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
		if((reg & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
		return reg;
	}



	/**
	 *  INC (HL) - Increment Memory Location
	 *
	 *  @param offset	   Memory Offset to Increment
	 */

	private final void inchl(int offset) {
		offset &= 0xFFFF;
		int location = mem_read(offset);
		int result = location + 1;
		result &= 0xFF;
		if ((result & 0x0F) == 0x00) F_ |= 0x10;
		else F_ &=~ 0x10;
		if (result == 0x80) F_ |= 0x04;
		else F_ &=~ 0x04;
		if (result == 0) F_ |= 0x40;
		else F_ &=~ 0x40;
		if ((result & 0x80) == 0) F_ &=~ 0x80;
		else F_ |= 0x80;
		F_ &=~ 0x02;
		if((result & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
		if((result & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
		mem_write(offset, result);
	}


	/**
	 *  DEC (HL) - Decrement Memory Location
	 *
	 *  @param offset	   Memory Offset to Increment
	 */

	private final void dechl(int offset) {
		offset &= 0xFFFF;
		int location = mem_read(offset);
		int result = location - 1;
		result &= 0xFF;
		if ((result & 0x0F) == 0x0F) F_ |= 0x10;
		else F_ &=~ 0x10;
		if (result == 0x7F) F_ |= 0x04;
		else F_ &=~ 0x04;
		if (result == 0) F_ |= 0x40;
		else F_ &=~ 0x40;
		if ((result & 0x80) == 0) F_ &=~ 0x80;
		else F_ |= 0x80;
		F_ |= 0x02;
		if((result & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
		if((result & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
		mem_write(offset, result);
	}



	/**
	 *  Execute DD/FD Prefixed Opcode
	 *
	 *  @param opcode	   Opcode hex value
	 */



	private final int doDDFD(int opcode, int Index_)
	{
		//int opcode = mem_read(++pc);
		int High_ = Index_ >> 8;
		int Low_ = Index_ & 0xFF;
		tstates -= opddstates[opcode];
		R_ = (R_&0x80) | ((R_+1)&0x7F);
		switch (opcode) {
				case 0x24: High_ = inc8(High_);		pc++; break;								// INC IXh *
				case 0x25: High_ = dec8(High_);		pc++; break;								// DEC IXh *
				case 0x26: High_ = mem_read(++pc);	pc++; break;								// LD IXh,n *
				case 0x2C: Low_ = inc8(Low_);		pc++; break;								// INC IXl *
				case 0x2D: Low_ = dec8(Low_);		pc++; break;								// DEC IXl *
				case 0x2E: Low_ = mem_read(++pc);	pc++; break;								// LD IXl,n
				case 0x44: B_ = (High_) & 0xFF;		pc++; break;								// LD B,IXh *
				case 0x45: B_ = (Low_) & 0xFF;		pc++; break;								// LD B,IXl *
				case 0x4C: C_ = (High_) & 0xFF;		pc++; break;								// LD C,IXh *
				case 0x4D: C_ = (Low_) & 0xFF;		pc++; break;								// LD C,IXl *
				case 0x54: D_ = (High_) & 0xFF;		pc++; break;								// LD D,IXh *
				case 0x55: D_ = (Low_) & 0xFF;		pc++; break;								// LD D,IXl *
				case 0x5C: E_ = (High_) & 0xFF;		pc++; break;								// LD E,IXh *
				case 0x5D: E_ = (Low_) & 0xFF;		pc++; break;								// LD E,IXl *
				case 0x60: High_ = (B_) & 0xFF;		pc++; break;								// LD IXh,B *
				case 0x61: High_ = (C_) & 0xFF;		pc++; break;								// LD IXh,C *
				case 0x62: High_ = (D_) & 0xFF;		pc++; break;								// LD IXh,D *
				case 0x63: High_ = (E_) & 0xFF;		pc++; break;								// LD IXh,E *
				case 0x64:							pc++; break;								// LD IXH,IXH*
				case 0x65: High_ = (Low_) & 0xFF;	pc++; break;								// LD IXH,IXL *
				case 0x67: High_ = (A_) & 0xFF;		pc++; break;								// LD IXh,A *
				case 0x68: Low_ = (B_) & 0xFF;		pc++; break;								// LD IXL,B *
				case 0x69: Low_ = (C_) & 0xFF;		pc++; break;								// LD IXl,C *
				case 0x6A: Low_ = (D_) & 0xFF;		pc++; break;								// LD IXL,D *
				case 0x6B: Low_ = (E_) & 0xFF;		pc++; break;								// LD IXl,E *
				case 0x6C: Low_ = (High_) & 0xFF;	pc++; break;								// LD IXl,IXh *
				case 0x6D:							pc++; break;								// LD IXl,IXl *
				case 0x6F: Low_ = (A_) & 0xFF;		pc++; break;								// LD IXl,A *
				case 0x7C: A_ = High_ & 0xFF;		pc++; break;								// LD A,IXh *
				case 0x7D: A_ = Low_ & 0xFF;		pc++; break;								// LD A,IXl *
				case 0x84: A_ = add(A_,High_);		pc++; break;								// ADD A,IXh *
				case 0x85: A_ = add(A_,Low_);		pc++; break;								// ADD A,IXl *
				case 0x8C: A_ = adc(A_,High_);		pc++; break;								// ADC A,IXH *
				case 0x8D: A_ = adc(A_,Low_);		pc++; break;								// ADC A,IXL *
				case 0x94: A_ = sub(A_,High_);		pc++; break;								// SUB IXh *
				case 0x95: A_ = sub(A_,Low_);		pc++; break;								// SUB IXl *
				case 0x9C: A_ = sbc(A_,High_);		pc++; break;								// SBC A,IXH *
				case 0x9D: A_ = sbc(A_,Low_);		pc++; break;								// SBC A,IXL *
				case 0xA4:
					{
						int value = High_;						// AND IXh *
						A_ &= value;
						if ((A_ & 0x80) == 0) F_ &=~ 0x80;
						else F_ |= 0x80;
						if (A_ == 0) F_ |= 0x40;
						else F_ &=~ 0x40;
						F_ |= 0x10;
						if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
						F_ &=~ 0x02;
						F_ &=~ 0x01;
						if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
						if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
					}
					pc++;
					break;
				case 0xA5:
					{
						int value = Low_;						// AND IXl *
						A_ &= value;
						if ((A_ & 0x80) == 0) F_ &=~ 0x80;
						else F_ |= 0x80;
						if (A_ == 0) F_ |= 0x40;
						else F_ &=~ 0x40;
						F_ |= 0x10;
						if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
						F_ &=~ 0x02;
						F_ &=~ 0x01;
						if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
						if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
					}
					pc++;
					break;
				case 0xAC:
					{
						int value = High_;						// XOR A IXH*
						A_ ^= value; // bitwise exclusive OR
						if (A_ == 0) F_ |= 0x40;
						else F_ &=~ 0x40;
						if ((A_ & 0x80) == 0) F_ &=~ 0x80;
						else F_ |= 0x80;
						F_ &=~ 0x10;
						F_ &=~ 0x02;
						F_ &=~ 0x01;
						if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
						if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
						if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
					}
					pc++;
					break;
				case 0xAD:
					{
						int value = Low_;						// XOR A IXL*
						A_ ^= value; // bitwise exclusive OR
						if (A_ == 0) F_ |= 0x40;
						else F_ &=~ 0x40;
						if ((A_ & 0x80) == 0) F_ &=~ 0x80;
						else F_ |= 0x80;
						F_ &=~ 0x10;
						F_ &=~ 0x02;
						F_ &=~ 0x01;
						if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
						if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
						if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
					}
					pc++;
					break;
				case 0xB4:
					{
						int value = High_;						 // OR A IXH*
						A_ |= value; // bitwise inclusive OR to turn relevant bits on
						if ((A_ & 0x80) == 0) F_ &=~ 0x80;
						else F_ |= 0x80;
						if (A_ == 0) F_ |= 0x40;
						else F_ &=~ 0x40;
						F_ &=~ 0x10;
						if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
						F_ &=~ 0x02;
						F_ &=~ 0x01;
						if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
						if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
					}
					pc++;
					break;
				case 0xB5:
					{
						int value = Low_;						 // OR A IXL*
						A_ |= value; // bitwise inclusive OR to turn relevant bits on
						if ((A_ & 0x80) == 0) F_ &=~ 0x80;
						else F_ |= 0x80;
						if (A_ == 0) F_ |= 0x40;
						else F_ &=~ 0x40;
						F_ &=~ 0x10;
						if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
						F_ &=~ 0x02;
						F_ &=~ 0x01;
						if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
						if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
					}
					pc++;
					break;
				case 0xBC:
					{
						int value = High_;					 // CP IXh *
						int result = A_ - value;
						int carrybits = A_ ^ value ^ result;
						if ((carrybits & 0x10) != 0 ) F_ |= 0x10;
						else F_ &=~ 0x10;
						if ((carrybits & 0x100) != 0) F_ |= 0x01;
						else F_ &=~ 0x01;
						if ((((carrybits << 1) ^ carrybits) & 0x100) != 0) F_ |= 0x04;
						else F_ &=~ 0x04;
						result &= 0xFF;
						if ((result & 0x80) == 0) F_ &=~ 0x80;
						else F_ |= 0x80;
						if (result == 0) F_ |= 0x40;
						else F_ &=~ 0x40;
						F_ |= 0x02;
						if((value & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
						if((value & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
					}
					pc++;
					break;
				case 0xBD:
					{
						int value = Low_;					 // CP IXl *
						int result = A_ - value;
						int carrybits = A_ ^ value ^ result;
						if ((carrybits & 0x10) != 0 ) F_ |= 0x10;
						else F_ &=~ 0x10;
						if ((carrybits & 0x100) != 0) F_ |= 0x01;
						else F_ &=~ 0x01;
						if ((((carrybits << 1) ^ carrybits) & 0x100) != 0) F_ |= 0x04;
						else F_ &=~ 0x04;
						result &= 0xFF;
						if ((result & 0x80) == 0) F_ &=~ 0x80;
						else F_ |= 0x80;
						if (result == 0) F_ |= 0x40;
						else F_ &=~ 0x40;
						F_ |= 0x02;
						if((value & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
						if((value & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
					}
					pc++;
					break;



			case 0x09:						// ADD IX,BC
				{
					int value = ((B_ << 8) | C_);
					int reg = ((High_ << 8) | Low_);
					int result = reg + value;
					int carrybits = reg ^ value ^ result;
					if ((carrybits & 0x1000) != 0 ) F_ |= 0x10;
					else F_ &=~ 0x10;
					if ((carrybits & 0x10000) != 0) F_ |= 0x01;
					else F_ &=~ 0x01;
					result &= 0xFFFF;
					Low_   = result & 0xFF;
					High_  = (result >> 8) & 0xFF;
					F_ &=~ 0x02;
					result >>= 8;			// Bits 3 and 5 affected by High_ byte addition
					if((result & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;}
					if((result & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;}
				}
				pc++;
				break;
			case 0x19:						// ADD IX,DE
				{
					int value = ((D_ << 8) | E_);
					int reg = ((High_ << 8) | Low_);
					int result = reg + value;
					int carrybits = reg ^ value ^ result;
					if ((carrybits & 0x1000) != 0 ) F_ |= 0x10;
					else F_ &=~ 0x10;
					if ((carrybits & 0x10000) != 0) F_ |= 0x01;
					else F_ &=~ 0x01;
					result &= 0xFFFF;
					Low_   = result & 0xFF;
					High_  = (result >> 8) & 0xFF;
					F_ &=~ 0x02;
					result >>= 8;			// Bits 3 and 5 affected by High_ byte addition
					if((result & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;}
					if((result & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;}
				}
				pc++;
				break;
			case 0x21:			 // LD IX,nn
				{
					int temp = (mem_read(pc + 2) << 8) + mem_read(pc + 1);
					Low_ = temp & 0xFF;
					High_ = (temp >> 8) & 0xFF;
				}
				pc += 3;
				break;
			case 0x22:
				mem_write((mem_read(pc + 2) << 8) + mem_read(pc + 1), Low_);		 // LD (nn),IX
				mem_write(((mem_read(pc + 2) << 8) + mem_read(pc + 1) + 1), High_);
				pc+=3;
				break;
			case 0x23:Low_ = (Low_ + 1) & 0xFF;if (Low_ == 0) High_ = (High_ + 1) & 0xFF; pc++; break;								// INC IX
			case 0x29:					 // ADD IX,IX
				{
					int value = ((High_ << 8) | Low_);
					int reg = ((High_ << 8) | Low_);
					int result = reg + value;
					int carrybits = reg ^ value ^ result;
					if ((carrybits & 0x1000) != 0 ) F_ |= 0x10;
					else F_ &=~ 0x10;
					if ((carrybits & 0x10000) != 0) F_ |= 0x01;
					else F_ &=~ 0x01;
					result &= 0xFFFF;
					Low_   = result & 0xFF;
					High_  = (result >> 8) & 0xFF;
					F_ &=~ 0x02;
					result >>= 8;			// Bits 3 and 5 affected by High_ byte addition
					if((result & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;}
					if((result & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;}
				}
				pc++;
				break;
			case 0x2A:
				Low_  = mem_read((mem_read(pc + 2) << 8) + mem_read(pc + 1));				// LD IX,(nn)
				High_ = mem_read((mem_read(pc + 2) << 8) + mem_read(pc + 1) + 1);
				pc+=3;
				break;
			case 0x2B:Low_ = (Low_ - 1) & 0xFF;if (Low_ == 255) High_ = (High_ - 1) & 0xFF; pc++; break;								// DEC IX
			case 0x34:inchl(((High_ << 8) | Low_) + mem_readSigned(++pc)); pc++; break;				 // INC (IX+d)
			case 0x35:dechl(((High_ << 8) | Low_) + mem_readSigned(++pc)); pc++; break;				 // DEC (IX+d)
			case 0x36:mem_write(((High_ << 8) | Low_) + mem_readSigned(++pc), mem_read(++pc)); pc++; break;// LD (IX+d), n
			case 0x39:						  // ADD IX,SP
				{
					int value = sp;
					int reg = ((High_ << 8) | Low_);
					int result = reg + value;
					int carrybits = reg ^ value ^ result;
					if ((carrybits & 0x1000) != 0 ) F_ |= 0x10;
					else F_ &=~ 0x10;
					if ((carrybits & 0x10000) != 0) F_ |= 0x01;
					else F_ &=~ 0x01;
					result &= 0xFFFF;
					Low_   = result & 0xFF;
					High_  = (result >> 8) & 0xFF;
					F_ &=~ 0x02;
					result >>= 8;			// Bits 3 and 5 affected by High_ byte addition
					if((result & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;}
					if((result & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;}
				}
				pc++;
				break;
			case 0x46: B_ = (mem_read(((High_ << 8) | Low_) + mem_readSigned(++pc)));	pc++; break;	// LD B,(IX+d)
			case 0x4E: C_ = (mem_read(((High_ << 8) | Low_) + mem_readSigned(++pc)));	pc++; break;	// LD C,(IX+d)
			case 0x56: D_ = (mem_read(((High_ << 8) | Low_) + mem_readSigned(++pc)));	pc++; break;	// LD D,(IX+d)
			case 0x5E: E_ = (mem_read(((High_ << 8) | Low_) + mem_readSigned(++pc)));	pc++; break;	// LD E,(IX+d)
			case 0x66: H_ = (mem_read(((High_ << 8) | Low_) + mem_readSigned(++pc)));	pc++; break;	// LD H,(IX+d)
			case 0x6E: L_ = (mem_read(((High_ << 8) | Low_) + mem_readSigned(++pc)));	pc++; break;	// LD L,(IX+d)
			case 0x70: mem_write(((High_ << 8) | Low_) + mem_readSigned(++pc), B_);		pc++; break;	// LD (IX+d),B
			case 0x71: mem_write(((High_ << 8) | Low_) + mem_readSigned(++pc), C_);		pc++; break;	// LD (IX+d),C
			case 0x72: mem_write(((High_ << 8) | Low_) + mem_readSigned(++pc), D_);		pc++; break;	// LD (IX+d),D
			case 0x73: mem_write(((High_ << 8) | Low_) + mem_readSigned(++pc), E_);		pc++; break;	// LD (IX+d),E
			case 0x74: mem_write(((High_ << 8) | Low_) + mem_readSigned(++pc), H_);		pc++; break;	// LD (IX+d),H
			case 0x75: mem_write(((High_ << 8) | Low_) + mem_readSigned(++pc), L_);		pc++; break;	// LD (IX+d),L
			case 0x77: mem_write(((High_ << 8) | Low_) + mem_readSigned(++pc), A_);		pc++; break;	// LD (IX+d),A
			case 0x7E: A_ = mem_read(((High_ << 8) | Low_) + mem_readSigned(++pc));		pc++; break;	// LD A,(IX+d)
			case 0x86: A_ = add(A_,mem_read(((High_ << 8) | Low_) + mem_readSigned(++pc)));	pc++; break;// ADD A,(IX+d)
			case 0x8E: A_ = adc(A_,mem_read(((High_ << 8) | Low_) + mem_readSigned(++pc)));	pc++; break;// ADC A,(IX+d)
			case 0x96: A_ = sub(A_,mem_read(((High_ << 8) | Low_) + mem_readSigned(++pc)));	pc++; break;// SUB A,(IX+d)
			case 0x9E: A_ = sbc(A_,mem_read(((High_ << 8) | Low_) + mem_readSigned(++pc)));	pc++; break;// SBC A,(IX+d)
			case 0xA6:
				{
					int value = mem_read(((High_ << 8) | Low_) + mem_readSigned(++pc));// pc++; break;	   // AND A,(IX+d)
					A_ &= value;
					if ((A_ & 0x80) == 0) F_ &=~ 0x80;
					else F_ |= 0x80;
					if (A_ == 0) F_ |= 0x40;
					else F_ &=~ 0x40;
					F_ |= 0x10;
					if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
					F_ &=~ 0x02;
					F_ &=~ 0x01;
					if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
					if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
				}
				pc++;
				break;
			case 0xAE:
				{
					int value = mem_read(((High_ << 8) | Low_) + mem_readSigned(++pc));// pc++; break;	   // XOR A,(IX+d)
					A_ ^= value; // bitwise exclusive OR
					if (A_ == 0) F_ |= 0x40;
					else F_ &=~ 0x40;
					if ((A_ & 0x80) == 0) F_ &=~ 0x80;
					else F_ |= 0x80;
					F_ &=~ 0x10;
					F_ &=~ 0x02;
					F_ &=~ 0x01;
					if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
					if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
					if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
				}
				pc++;
				break;
			case 0xB6:
				{
					int value = mem_read(((High_ << 8) | Low_) + mem_readSigned(++pc));// pc++; break;			// OR A,(IX+d)
					A_ |= value; // bitwise inclusive OR to turn relevant bits on
					if ((A_ & 0x80) == 0) F_ &=~ 0x80;
					else F_ |= 0x80;
					if (A_ == 0) F_ |= 0x40;
					else F_ &=~ 0x40;
					F_ &=~ 0x10;
					if(PARITY[A_ & 0xFF]){F_ |= 0x04;}else{F_ &=~ 0x04;};
					F_ &=~ 0x02;
					F_ &=~ 0x01;
					if((A_ & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
					if((A_ & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
				}
				pc++;
				break;
			case 0xBE:
				{
					int value = mem_read(((High_ << 8) | Low_) + mem_readSigned(++pc));// pc++; break;			// CP (IX+d)
					int result = A_ - value;
					int carrybits = A_ ^ value ^ result;
					if ((carrybits & 0x10) != 0 ) F_ |= 0x10;
					else F_ &=~ 0x10;
					if ((carrybits & 0x100) != 0) F_ |= 0x01;
					else F_ &=~ 0x01;
					if ((((carrybits << 1) ^ carrybits) & 0x100) != 0) F_ |= 0x04;
					else F_ &=~ 0x04;
					result &= 0xFF;
					if ((result & 0x80) == 0) F_ &=~ 0x80;
					else F_ |= 0x80;
					if (result == 0) F_ |= 0x40;
					else F_ &=~ 0x40;
					F_ |= 0x02;
					if((value & 0x08) == 0x08){F_ |= 0x08;}else{F_ &=~ 0x08;};
					if((value & 0x20) == 0x20){F_ |= 0x20;}else{F_ &=~ 0x20;};
				}
				pc++;
				break;
			case 0xCB: 						 // DDCB/DDFD Opcode
				{
					int location = (((High_ << 8) | Low_) + mem_readSigned(++pc)) & 0xFFFF;
					int opcodeCB = mem_read(++pc);
					tstates -= 23;
					switch (opcodeCB) {
						case 0x06: mem_write(location, rlc(mem_read(location))); break;			// RLC (IX)
						case 0x0E: mem_write(location, rrc(mem_read(location)));	break;		// RRC (IX)
						case 0x16: mem_write(location, rl(mem_read(location))); break;			// RL (IX)
						case 0x1E: mem_write(location, rr(mem_read(location))); break;			// RR (IX)
						case 0x26: mem_write(location, sla(mem_read(location))); break;			// SLA (IX)
						case 0x2E: mem_write(location, sra(mem_read(location))); break;			// SRA (IX)
						case 0x36: mem_write(location, sll(mem_read(location))); break;			// SLL (IX)
						case 0x3E: mem_write(location, srl(mem_read(location))); break;			// SRL (IX)
						case 0x46: bit(mem_read(location), 0); tstates += 3; break;				// BIT 0,(IX)
						case 0x4E: bit(mem_read(location), 1); tstates += 3; break;				// BIT 1,(IX)
						case 0x56: bit(mem_read(location), 2); tstates += 3; break;				// BIT 2,(IX)
						case 0x5E: bit(mem_read(location), 3); tstates += 3; break;				// BIT 3,(IX)
						case 0x66: bit(mem_read(location), 4); tstates += 3; break;				// BIT 4,(IX)
						case 0x6E: bit(mem_read(location), 5); tstates += 3; break;				// BIT 5,(IX)
						case 0x76: bit(mem_read(location), 6); tstates += 3; break;				// BIT 6,(IX)
						case 0x7E: bit(mem_read(location), 7); tstates += 3; break;				// BIT 7,(IX)
						case 0x86: mem_write(location, res(mem_read(location), 0)); break;		// RES 0,(IX)
						case 0x8E: mem_write(location, res(mem_read(location), 1)); break;		// RES 1,(IX)
						case 0x96: mem_write(location, res(mem_read(location), 2)); break;		// RES 2,(IX)
						case 0x9E: mem_write(location, res(mem_read(location), 3)); break;		// RES 3,(IX)
						case 0xA6: mem_write(location, res(mem_read(location), 4)); break;		// RES 4,(IX)
						case 0xAE: mem_write(location, res(mem_read(location), 5)); break;		// RES 5,(IX)
						case 0xB6: mem_write(location, res(mem_read(location), 6)); break;		// RES 6,(IX)
						case 0xBE: mem_write(location, res(mem_read(location), 7)); break;		// RES 7,(IX)
						case 0xC6: mem_write(location, setBit(mem_read(location), 0)); break;	// SET 0,(IX)
						case 0xCE: mem_write(location, setBit(mem_read(location), 1)); break;	// SET 1,(IX)
						case 0xD6: mem_write(location, setBit(mem_read(location), 2)); break;	// SET 2,(IX)
						case 0xDE: mem_write(location, setBit(mem_read(location), 3)); break;	// SET 3,(IX)
						case 0xE6: mem_write(location, setBit(mem_read(location), 4)); break;	// SET 4,(IX)
						case 0xEE: mem_write(location, setBit(mem_read(location), 5)); break;	// SET 5,(IX)
						case 0xF6: mem_write(location, setBit(mem_read(location), 6)); break;	// SET 6,(IX)
						case 0xFE: mem_write(location, setBit(mem_read(location), 7)); break;	// SET 7,(IX)
						default:
							tstates += 23;
							System.out.println("Unimplemented DDCB/FDCB Opcode : 0x" + Integer.toHexString(opcodeCB & 0xFF).toUpperCase());
							break;
					} // end of switch
				}
				pc++;
				break;
			case 0xE1:					   // POP IX
				Low_ = mem_read(sp);
				High_ = mem_read(sp + 1);
				sp += 2;
				sp &= 0xFFFF;
				pc++;
				break;
			case 0xE3:	 // EX SP,(IX)
				{
					int temp = ((High_ << 8) | Low_);
					Low_  = mem_read(sp);
					High_ = mem_read(sp + 1);
					mem_write(sp, temp & 0xFF);
					mem_write(sp+1, temp>>8);
				}
				pc++;
				break;
			case 0xE5:
				mem_write(--sp, High_);mem_write(--sp, Low_);
				pc++;
				break;			// PUSH IX
			case 0xE9: pc = ((High_ << 8) | Low_);					  break;			// JP (IX)
			case 0xF9: sp = ((High_ << 8) | Low_);				pc++; break;			// LD SP,IX
			// Unimplented DD/FD Opcode
			default:
				System.out.println("Unimplemented DD/FD Opcode : 0x" + Integer.toHexString(opcode & 0xFF).toUpperCase());
				//pc++;
				break;
		} // end of switch
		Index_  = (High_ << 8) + Low_;
		return Index_;
	}



	/**
	 *  Execute CB Prefixed Opcode
	 *
	 *  @param opcode	   Opcode hex value
	 */

	private final void doCB(int opcode)
	{
		//opcode = mem_read(++pc);
		R_ = (R_&0x80) | ((R_+1)&0x7F);
		tstates -= opcbstates[opcode];
		int HW = (H_ << 8) | L_;
		int HL = mem_read(HW);
		switch(opcode)
		{
			case 0x00: B_ = rlc(B_) & 0xFF; break;				// RLC B
			case 0x01: C_ = rlc(C_) & 0xFF; break;				// RLC C
			case 0x02: D_ = rlc(D_) & 0xFF; break;				// RLC D
			case 0x03: E_ = rlc(E_) & 0xFF; break;				// RLC E
			case 0x04: H_ = rlc(H_) & 0xFF; break;				// RLC H
			case 0x05: L_ = rlc(L_) & 0xFF; break;				// RLC L
			case 0x06: mem_write(HW, rlc(HL)); break;			// RLC (HL)
			case 0x07: A_ = rlc(A_) & 0xFF; break;				// RLC A
			case 0x08: B_ = rrc(B_) & 0xFF; break;				// RRC B
			case 0x09: C_ = rrc(C_) & 0xFF; break;				// RRC C
			case 0x0A: D_ = rrc(D_) & 0xFF; break;				// RRC D
			case 0x0B: E_ = rrc(E_) & 0xFF; break;				// RRC E
			case 0x0C: H_ = rrc(H_) & 0xFF; break;				// RRC H
			case 0x0D: L_ = rrc(L_) & 0xFF; break;				// RRC L
			case 0x0E: mem_write(HW, rrc(HL)); break;			// RRC (HL)
			case 0x0F: A_ = rrc(A_) & 0xFF; break;				// RRC A
			case 0x10: B_ = rl(B_) & 0xFF; break;				// RL B
			case 0x11: C_ = rl(C_) & 0xFF; break;				// RL C
			case 0x12: D_ = rl(D_) & 0xFF; break;				// RL D
			case 0x13: E_ = rl(E_) & 0xFF; break;				// RL E
			case 0x14: H_ = rl(H_) & 0xFF; break;				// RL H
			case 0x15: L_ = rl(L_) & 0xFF; break;				// RL L
			case 0x16: mem_write(HW, rl(HL)); break;			// RL (HL)
			case 0x17: A_ = rl(A_) & 0xFF; break;				// RL A
			case 0x18: B_ = rr(B_) & 0xFF; break;				// RR B
			case 0x19: C_ = rr(C_) & 0xFF; break;				// RR C
			case 0x1A: D_ = rr(D_) & 0xFF; break;				// RR D
			case 0x1B: E_ = rr(E_) & 0xFF; break;				// RR E
			case 0x1C: H_ = rr(H_) & 0xFF; break;				// RR H
			case 0x1D: L_ = rr(L_) & 0xFF; break;				// RR L
			case 0x1E: mem_write(HW, rr(HL)); break;			// RR (HL)
			case 0x1F: A_ = rr(A_) & 0xFF; break;				// RR A
			case 0x20: B_ = sla(B_) & 0xFF; break;				// SLA B
			case 0x21: C_ = sla(C_) & 0xFF; break;				// SLA C
			case 0x22: D_ = sla(D_) & 0xFF; break;				// SLA D
			case 0x23: E_ = sla(E_) & 0xFF; break;				// SLA E
			case 0x24: H_ = sla(H_) & 0xFF; break;				// SLA H
			case 0x25: L_ = sla(L_) & 0xFF; break;				// SLA L
			case 0x26: mem_write(HW, sla(HL)); break;			// SLA (HL)
			case 0x27: A_ = sla(A_) & 0xFF; break;				// SLA A
			case 0x28: B_ = sra(B_) & 0xFF; break;				// SRA B
			case 0x29: C_ = sra(C_) & 0xFF; break;				// SRA C
			case 0x2A: D_ = sra(D_) & 0xFF; break;				// SRA D
			case 0x2B: E_ = sra(E_) & 0xFF; break;				// SRA E
			case 0x2C: H_ = sra(H_) & 0xFF; break;				// SRA H
			case 0x2D: L_ = sra(L_) & 0xFF; break;				// SRA L
			case 0x2E: mem_write(HW, sra(HL)); break;			// SRA (HL)
			case 0x2F: A_ = sra(A_) & 0xFF; break;				// SRA A
			case 0x30: B_ = sll(B_) & 0xFF; break;				// SLL B
			case 0x31: C_ = sll(C_) & 0xFF; break;				// SLL C
			case 0x32: D_ = sll(D_) & 0xFF; break;				// SLL D
			case 0x33: E_ = sll(E_) & 0xFF; break;				// SLL E
			case 0x34: H_ = sll(H_) & 0xFF; break;				// SLL H
			case 0x35: L_ = sll(L_) & 0xFF; break;				// SLL L
			case 0x36: mem_write(HW, sll(HL)); break;			// SLL (HL)
			case 0x37: A_ = sll(A_) & 0xFF; break;				// SLL A
			case 0x38: B_ = srl(B_) & 0xFF; break;				// SRL B
			case 0x39: C_ = srl(C_) & 0xFF; break;				// SRL C
			case 0x3A: D_ = srl(D_) & 0xFF; break;				// SRL D
			case 0x3B: E_ = srl(E_) & 0xFF; break;				// SRL E
			case 0x3C: H_ = srl(H_) & 0xFF; break;				// SRL H
			case 0x3D: L_ = srl(L_) & 0xFF; break;				// SRL L
			case 0x3E: mem_write(HW, srl(HL)); break;			// SRL (HL)
			case 0x3F: A_ = srl(A_) & 0xFF; break;				// SRL A
			case 0x40: bit(B_, 0); break;						// BIT 0,B
			case 0x41: bit(C_, 0); break;						// BIT 0,C
			case 0x42: bit(D_, 0); break;						// BIT 0,D
			case 0x43: bit(E_, 0); break;						// BIT 0,E
			case 0x44: bit(H_, 0); break;						// BIT 0,H
			case 0x45: bit(L_, 0); break;						// BIT 0,L
			case 0x46: bit(HL, 0); break;						// BIT 0,(HL)
			case 0x47: bit(A_, 0); break;						// BIT 0,A
			case 0x48: bit(B_, 1); break;						// BIT 1,B
			case 0x49: bit(C_, 1); break;						// BIT 1,C
			case 0x4A: bit(D_, 1); break;						// BIT 1,D
			case 0x4B: bit(E_, 1); break;						// BIT 1,E
			case 0x4C: bit(H_, 1); break;						// BIT 1,H
			case 0x4D: bit(L_, 1); break;						// BIT 1,L
			case 0x4E: bit(HL, 1); break;						// BIT 1,(HL)
			case 0x4F: bit(A_, 1); break;						// BIT 1,A
			case 0x50: bit(B_, 2); break;						// BIT 2,B
			case 0x51: bit(C_, 2); break;						// BIT 2,C
			case 0x52: bit(D_, 2); break;						// BIT 2,D
			case 0x53: bit(E_, 2); break;						// BIT 2,E
			case 0x54: bit(H_, 2); break;						// BIT 2,H
			case 0x55: bit(L_, 2); break;						// BIT 2,L
			case 0x56: bit(HL, 2); break;						// BIT 2,(HL)
			case 0x57: bit(A_, 2); break;						// BIT 2,A
			case 0x58: bit(B_, 3); break;						// BIT 3,B
			case 0x59: bit(C_, 3); break;						// BIT 3,C
			case 0x5A: bit(D_, 3); break;						// BIT 3,D
			case 0x5B: bit(E_, 3); break;						// BIT 3,E
			case 0x5C: bit(H_, 3); break;						// BIT 3,H
			case 0x5D: bit(L_, 3); break;						// BIT 3,L
			case 0x5E: bit(HL, 3); break;						// BIT 3,(HL)
			case 0x5F: bit(A_, 3); break;						// BIT 3,A
			case 0x60: bit(B_, 4); break;						// BIT 4,B
			case 0x61: bit(C_, 4); break;						// BIT 4,C
			case 0x62: bit(D_, 4); break;						// BIT 4,D
			case 0x63: bit(E_, 4); break;						// BIT 4,E
			case 0x64: bit(H_, 4); break;						// BIT 4,H
			case 0x65: bit(L_, 4); break;						// BIT 4,L
			case 0x66: bit(HL, 4); break;						// BIT 4,(HL)
			case 0x67: bit(A_, 4); break;						// BIT 4,A
			case 0x68: bit(B_, 5); break;						// BIT 5,B
			case 0x69: bit(C_, 5); break;						// BIT 5,C
			case 0x6A: bit(D_, 5); break;						// BIT 5,D
			case 0x6B: bit(E_, 5); break;						// BIT 5,E
			case 0x6C: bit(H_, 5); break;						// BIT 5,H
			case 0x6D: bit(L_, 5); break;						// BIT 5,L
			case 0x6E: bit(HL, 5); break;						// BIT 5,(HL)
			case 0x6F: bit(A_, 5); break;						// BIT 5,A
			case 0x70: bit(B_, 6); break;						// BIT 6,B
			case 0x71: bit(C_, 6); break;						// BIT 6,C
			case 0x72: bit(D_, 6); break;						// BIT 6,D
			case 0x73: bit(E_, 6); break;						// BIT 6,E
			case 0x74: bit(H_, 6); break;						// BIT 6,H
			case 0x75: bit(L_, 6); break;						// BIT 6,L
			case 0x76: bit(HL, 6); break;						// BIT 6,(HL)
			case 0x77: bit(A_, 6); break;						// BIT 6,A
			case 0x78: bit(B_, 7); break;						// BIT 7,B
			case 0x79: bit(C_, 7); break;						// BIT 7,C
			case 0x7A: bit(D_, 7); break;						// BIT 7,D
			case 0x7B: bit(E_, 7); break;						// BIT 7,E
			case 0x7C: bit(H_, 7); break;						// BIT 7,H
			case 0x7D: bit(L_, 7); break;						// BIT 7,L
			case 0x7E: bit(HL, 7); break;						// BIT 7,(HL)
			case 0x7F: bit(A_, 7); break;						// BIT 7,A
			case 0x80: B_ = res(B_, 0) & 0xFF; break;			// RES 0,B
			case 0x81: C_ = res(C_, 0) & 0xFF; break;			// RES 0,C
			case 0x82: D_ = res(D_, 0) & 0xFF; break;			// RES 0,D
			case 0x83: E_ = res(E_, 0) & 0xFF; break;			// RES 0,E
			case 0x84: H_ = res(H_, 0) & 0xFF; break;			// RES 0,H
			case 0x85: L_ = res(L_, 0) & 0xFF; break;			// RES 0,L
			case 0x86: mem_write(HW, res(HL, 0)); break;		// RES 0,(HL)
			case 0x87: A_ = res(A_, 0) & 0xFF; break;			// RES 0,A
			case 0x88: B_ = res(B_, 1) & 0xFF; break;			// RES 1,B
			case 0x89: C_ = res(C_, 1) & 0xFF; break;			// RES 1,C
			case 0x8A: D_ = res(D_, 1) & 0xFF; break;			// RES 1,D
			case 0x8B: E_ = res(E_, 1) & 0xFF; break;			// RES 1,E
			case 0x8C: H_ = res(H_, 1) & 0xFF; break;			// RES 1,H
			case 0x8D: L_ = res(L_, 1) & 0xFF; break;			// RES 1,L
			case 0x8E: mem_write(HW, res(HL, 1)); break;		// RES 1,(HL)
			case 0x8F: A_ = res(A_, 1) & 0xFF; break;			// RES 1,A
			case 0x90: B_ = res(B_, 2) & 0xFF; break;			// RES 2,B
			case 0x91: C_ = res(C_, 2) & 0xFF; break;			// RES 2,C
			case 0x92: D_ = res(D_, 2) & 0xFF; break;			// RES 2,D
			case 0x93: E_ = res(E_, 2) & 0xFF; break;			// RES 2,E
			case 0x94: H_ = res(H_, 2) & 0xFF; break;			// RES 2,H
			case 0x95: L_ = res(L_, 2) & 0xFF; break;			// RES 2,L
			case 0x96: mem_write(HW, res(HL, 2)); break;		// RES 2,(HL)
			case 0x97: A_ = res(A_, 2) & 0xFF; break;			// RES 2,A
			case 0x98: B_ = res(B_, 3) & 0xFF; break;			// RES 3,B
			case 0x99: C_ = res(C_, 3) & 0xFF; break;			// RES 3,C
			case 0x9A: D_ = res(D_, 3) & 0xFF; break;			// RES 3,D
			case 0x9B: E_ = res(E_, 3) & 0xFF; break;			// RES 3,E
			case 0x9C: H_ = res(H_, 3) & 0xFF; break;			// RES 3,H
			case 0x9D: L_ = res(L_, 3) & 0xFF; break;			// RES 3,L
			case 0x9E: mem_write(HW, res(HL, 3)); break;		// RES 3,(HL)
			case 0x9F: A_ = res(A_, 3) & 0xFF; break;			// RES 3,A
			case 0xA0: B_ = res(B_, 4) & 0xFF; break;			// RES 4,B
			case 0xA1: C_ = res(C_, 4) & 0xFF; break;			// RES 4,C
			case 0xA2: D_ = res(D_, 4) & 0xFF; break;			// RES 4,D
			case 0xA3: E_ = res(E_, 4) & 0xFF; break;			// RES 4,E
			case 0xA4: H_ = res(H_, 4) & 0xFF; break;			// RES 4,H
			case 0xA5: L_ = res(L_, 4) & 0xFF; break;			// RES 4,L
			case 0xA6: mem_write(HW, res(HL, 4)); break;  		// RES 4,(HL)
			case 0xA7: A_ = res(A_, 4) & 0xFF; break;			// RES 4,A
			case 0xA8: B_ = res(B_, 5) & 0xFF; break;			// RES 5,B
			case 0xA9: C_ = res(C_, 5) & 0xFF; break;			// RES 5,C
			case 0xAA: D_ = res(D_, 5) & 0xFF; break;			// RES 5,D
			case 0xAB: E_ = res(E_, 5) & 0xFF; break;			// RES 5,E
			case 0xAC: H_ = res(H_, 5) & 0xFF; break;			// RES 5,H
			case 0xAD: L_ = res(L_, 5) & 0xFF; break;			// RES 5,L
			case 0xAE: mem_write(HW, res(HL, 5)); break;		// RES 5,(HL)
			case 0xAF: A_ = res(A_, 5) & 0xFF; break;			// RES 5,A
			case 0xB0: B_ = res(B_, 6) & 0xFF; break;			// RES 6,B
			case 0xB1: C_ = res(C_, 6) & 0xFF; break;			// RES 6,C
			case 0xB2: D_ = res(D_, 6) & 0xFF; break;			// RES 6,D
			case 0xB3: E_ = res(E_, 6) & 0xFF; break;			// RES 6,E
			case 0xB4: H_ = res(H_, 6) & 0xFF; break;			// RES 6,H
			case 0xB5: L_ = res(L_, 6) & 0xFF; break;			// RES 6,L
			case 0xB6: mem_write(HW, res(HL, 6)); break;		// RES 6,(HL)
			case 0xB7: A_ = res(A_, 6) & 0xFF; break;			// RES 6,A
			case 0xB8: B_ = res(B_, 7) & 0xFF; break;			// RES 7,B
			case 0xB9: C_ = res(C_, 7) & 0xFF; break;			// RES 7,C
			case 0xBA: D_ = res(D_, 7) & 0xFF; break;			// RES 7,D
			case 0xBB: E_ = res(E_, 7) & 0xFF; break;			// RES 7,E
			case 0xBC: H_ = res(H_, 7) & 0xFF; break;			// RES 7,H
			case 0xBD: L_ = res(L_, 7) & 0xFF; break;			// RES 7,L
			case 0xBE: mem_write(HW, res(HL, 7)); break;		// RES 7,(HL)
			case 0xBF: A_ = res(A_, 7) & 0xFF; break;			// RES 7,A
			case 0xC0: B_ = setBit(B_, 0) & 0xFF; break;		// SET 0,B
			case 0xC1: C_ = setBit(C_, 0) & 0xFF; break;		// SET 0,C
			case 0xC2: D_ = setBit(D_, 0) & 0xFF; break;		// SET 0,D
			case 0xC3: E_ = setBit(E_, 0) & 0xFF; break;		// SET 0,E
			case 0xC4: H_ = setBit(H_, 0) & 0xFF; break;		// SET 0,H
			case 0xC5: L_ = setBit(L_, 0) & 0xFF; break;		// SET 0,L
			case 0xC6: mem_write(HW, setBit(HL, 0)); break;		// SET 0,(HL)
			case 0xC7: A_ = setBit(A_, 0) & 0xFF; break;		// SET 0,A
			case 0xC8: B_ = setBit(B_, 1) & 0xFF; break;		// SET 1,B
			case 0xC9: C_ = setBit(C_, 1) & 0xFF; break;		// SET 1,C
			case 0xCA: D_ = setBit(D_, 1) & 0xFF; break;		// SET 1,D
			case 0xCB: E_ = setBit(E_, 1) & 0xFF; break;		// SET 1,E
			case 0xCC: H_ = setBit(H_, 1) & 0xFF; break;		// SET 1,H
			case 0xCD: L_ = setBit(L_, 1) & 0xFF; break;		// SET 1,L
			case 0xCE: mem_write(HW, setBit(HL, 1)); break;		// SET 1,(HL)
			case 0xCF: A_ = setBit(A_, 1) & 0xFF; break;		// SET 1,A
			case 0xD0: B_ = setBit(B_, 2) & 0xFF; break;		// SET 2,B
			case 0xD1: C_ = setBit(C_, 2) & 0xFF; break;		// SET 2,C
			case 0xD2: D_ = setBit(D_, 2) & 0xFF; break;		// SET 2,D
			case 0xD3: E_ = setBit(E_, 2) & 0xFF; break;		// SET 2,E
			case 0xD4: H_ = setBit(H_, 2) & 0xFF; break;		// SET 2,H
			case 0xD5: L_ = setBit(L_, 2) & 0xFF; break;		// SET 2,L
			case 0xD6: mem_write(HW, setBit(HL, 2)); break;		// SET 2,(HL)
			case 0xD7: A_ = setBit(A_, 2) & 0xFF; break;		// SET 2,A
			case 0xD8: B_ = setBit(B_, 3) & 0xFF; break;		// SET 3,B
			case 0xD9: C_ = setBit(C_, 3) & 0xFF; break;		// SET 3,C
			case 0xDA: D_ = setBit(D_, 3) & 0xFF; break;		// SET 3,D
			case 0xDB: E_ = setBit(E_, 3) & 0xFF; break;		// SET 3,E
			case 0xDC: H_ = setBit(H_, 3) & 0xFF; break;		// SET 3,H
			case 0xDD: L_ = setBit(L_, 3) & 0xFF; break;		// SET 3,L
			case 0xDE: mem_write(HW, setBit(HL, 3)); break;		// SET 3,(HL)
			case 0xDF: A_ = setBit(A_, 3) & 0xFF; break;		// SET 3,A
			case 0xE0: B_ = setBit(B_, 4) & 0xFF; break;		// SET 4,B
			case 0xE1: C_ = setBit(C_, 4) & 0xFF; break;		// SET 4,C
			case 0xE2: D_ = setBit(D_, 4) & 0xFF; break;		// SET 4,D
			case 0xE3: E_ = setBit(E_, 4) & 0xFF; break;		// SET 4,E
			case 0xE4: H_ = setBit(H_, 4) & 0xFF; break;		// SET 4,H
			case 0xE5: L_ = setBit(L_, 4) & 0xFF; break;		// SET 4,L
			case 0xE6: mem_write(HW, setBit(HL, 4)); break;		// SET 4,(HL)
			case 0xE7: A_ = setBit(A_, 4) & 0xFF; break;		// SET 4,A
			case 0xE8: B_ = setBit(B_, 5) & 0xFF; break;		// SET 5,B
			case 0xE9: C_ = setBit(C_, 5) & 0xFF; break;		// SET 5,C
			case 0xEA: D_ = setBit(D_, 5) & 0xFF; break;		// SET 5,D
			case 0xEB: E_ = setBit(E_, 5) & 0xFF; break;		// SET 5,E
			case 0xEC: H_ = setBit(H_, 5) & 0xFF; break;		// SET 5,H
			case 0xED: L_ = setBit(L_, 5) & 0xFF; break;		// SET 5,L
			case 0xEE: mem_write(HW, setBit(HL, 5)); break;		// SET 5,(HL)
			case 0xEF: A_ = setBit(A_, 5) & 0xFF; break; 		// SET 5,A
			case 0xF0: B_ = setBit(B_, 6) & 0xFF; break;		// SET 6,B
			case 0xF1: C_ = setBit(C_, 6) & 0xFF; break;		// SET 6,C
			case 0xF2: D_ = setBit(D_, 6) & 0xFF; break;		// SET 6,D
			case 0xF3: E_ = setBit(E_, 6) & 0xFF; break;		// SET 6,E
			case 0xF4: H_ = setBit(H_, 6) & 0xFF; break;		// SET 6,H
			case 0xF5: L_ = setBit(L_, 6) & 0xFF; break;		// SET 6,L
			case 0xF6: mem_write(HW, setBit(HL, 6)); break;		// SET 6,(HL)
			case 0xF7: A_ = setBit(A_, 6) & 0xFF; break;		// SET 6,A
			case 0xF8: B_ = setBit(B_, 7) & 0xFF; break;		// SET 7,B
			case 0xF9: C_ = setBit(C_, 7) & 0xFF; break;		// SET 7,C
			case 0xFA: D_ = setBit(D_, 7) & 0xFF; break;		// SET 7,D
			case 0xFB: E_ = setBit(E_, 7) & 0xFF; break;		// SET 7,E
			case 0xFC: H_ = setBit(H_, 7) & 0xFF; break;		// SET 7,H
			case 0xFD: L_ = setBit(L_, 7) & 0xFF; break;		// SET 7,L
			case 0xFE: mem_write(HW, setBit(HL, 7)); break;		// SET 7,(HL)
			case 0xFF: A_ = setBit(A_, 7) & 0xFF; break;		// SET 7,A
			// Unimplented CB Opcode
			default:
				System.out.println("Unimplemented CB Opcode : 0x"+Integer.toHexString(opcode).toUpperCase());
				break;
		}
		pc++;		   // All CB opcodes are one byte (CB n).
	}



	/**
	 *  CB RLC - Rotate Left Carry
	 *
	 *  @param value		Value to adjust
	 *
	 *  @return			 Adjusted value
	 */

	private final int rlc(int value){
		
		int temp = value;
		value = (value << 1) & 0xFF;
		if ((temp & 0x80) == 0x80) {
			F_ |= 0x01;
			value |= 0x01;
		} else {
			F_ &=~ 0x01;
		}
		if (value == 0) F_ |= 0x40;
		else F_ &=~ 0x40;
		if ((value & 0x80) != 0) F_ |= 0x80;
		else F_ &=~ 0x80;
		F_&=~ 0x02;
		if(PARITY[value & 0xFF]) F_ |= 0x04;
		else F_ &=~ 0x04;
		F_ &=~ 0x10;
		return value;
	}


	/**
	 *  CB RRC - Rotate Right Carry
	 *
	 *  @param value		Value to adjust
	 *
	 *  @return			 Adjusted value
	 */

	private final int rrc(int value) {
		int temp = value;
		value >>= 1;
		if ((temp & 0x01) == 0x01) {
			F_ |= 0x01;
			value |= 0x80;
		} else {
			F_ &=~ 0x01;
		}
		if (value == 0) F_ |= 0x40;
		else F_ &=~ 0x40;
		if ((value & 0x80) != 0) F_ |= 0x80;
		else F_ &=~ 0x80;
		F_ &=~ 0x02;
		if(PARITY[value & 0xFF]) F_ |= 0x04;
		else F_ &=~ 0x04;
		F_ &=~ 0x10;
		return value;
	}


	/**
	 *  CB RL - Rotate Left
	 *
	 *  @param value		Value to adjust
	 *
	 *  @return			 Adjusted value
	 */

	private final int rl(int value){
		int temp = value;
		value = (value << 1) & 0xFF;
		if (((F_ & 0x01) == 0x01)) value |= 0x01;
		if ((temp & 0x80) == 0x80) F_ |= 0x01;
		else F_ &=~ 0x01;
		if (value == 0) F_ |= 0x40;
		else F_ &=~ 0x40;
		if ((value & 0x80) != 0) F_ |= 0x80;
		else F_ &=~ 0x80;
		F_ &=~ 0x02;
		if(PARITY[value & 0xFF]) F_ |= 0x04;
		else F_ &=~ 0x04;
		F_ &=~ 0x10;
		return value;
	}


	/**
	 *  CB RR - Rotate Right
	 *
	 *  @param value		Value to adjust
	 *
	 *  @return			 Adjusted value
	 */

	private final int rr(int value){
		int temp = value;
		value >>= 1;
		if (((F_ & 0x01) == 0x01)) value |= 0x80;
		if ((temp & 0x01) != 0) F_ |= 0x01;
		else F_ &=~ 0x01;
		if (value == 0) F_ |= 0x40;
		else F_ &=~ 0x40;
		if ((value & 0x80) != 0) F_ |= 0x80;
		else F_ &=~ 0x80;
		F_ &=~ 0x02;
		if(PARITY[value & 0xFF]) F_ |= 0x04;
		else F_ &=~ 0x04;
		F_ &=~ 0x10;
		return value;
	}


	/**
	 *  CB SLA - Shift Left Arithmetic
	 *
	 *  @param value		Value to adjust
	 *
	 *  @return			 Adjusted value
	 */

	private final int sla(int value) {
		if ((value & 0x80) == 0x80) F_ |= 0x01;
		else F_ &=~ 0x01;
		value = (value<<1) & 0xFF;
		value &= ~0x01;
		if (value == 0) F_ |= 0x40;
		else F_ &=~ 0x40;
		if ((value & 0x80) != 0) F_ |= 0x80;
		else F_ &=~ 0x80;
		F_ &=~ 0x02;
		if(PARITY[value & 0xFF]) F_ |= 0x04;
		else F_ &=~ 0x04;
		F_ &=~ 0x10;
		return value;
	}


	/**
	 *  CB SRA - Shift Right Arithmetic
	 *
	 *  @param value		Value to adjust
	 *
	 *  @return			 Adjusted value
	 */

	private final int sra(int value){
		if ((value & 0x01) == 0x01) F_ |= 0x01;
		else F_ &=~ 0x01;
		int temp = (value & 0x80);
		value = (value >> 1) |temp;
		if (value == 0) F_ |= 0x40;
		else F_ &=~ 0x40;
		if ((value & 0x80) != 0) F_ |= 0x80;
		else F_ &=~ 0x80;
		F_ &=~ 0x02;
		if(PARITY[value & 0xFF]) F_ |= 0x04;
		else F_ &=~ 0x04;
		F_ &=~ 0x10;
		return value;
	}

	/**
	 *  CB SLL - Logical Left Shift
	 *
	 *  @param value		Value to adjust
	 *
	 *  @return			 Adjusted value
	 */

	private final int sll(int value){
		if ((value & 0x80) == 0x80) F_ |= 0x01;
		else F_ &=~ 0x01;
		value = (value<<1) & 0xFF;
		value |= 0x01;
		if (value == 0) F_ |= 0x40;
		else F_ &=~ 0x40;
		if ((value & 0x80) != 0) F_ |= 0x80;
		else F_ &=~ 0x80;
		F_ &=~ 0x02;
		if(PARITY[value & 0xFF]) F_ |= 0x04;
		else F_ &=~ 0x04;
		F_ &=~ 0x10;
		return value;
	}

	/**
	 *  CB SRL - Logical Shift Right
	 *
	 *  @param value		Value to adjust
	 *
	 *  @return			 Adjusted value
	 */

	private final int srl(int value){
		if ((value & 0x01) == 0x01) F_ |= 0x01;
		else F_ &=~ 0x01;
		value >>= 1;
		value &= ~0x80;
		if (value == 0) F_ |= 0x40;
		else F_ &=~ 0x40;
		if ((value & 0x80) == 0x80) F_ |= 0x80;
		else F_ &=~ 0x80;
		F_ &=~ 0x02;
		if(PARITY[value & 0xFF]) F_ |= 0x04;
		else F_ &=~ 0x04;
		F_ &=~ 0x10;
		return value;
	}


	/**
	 *  CB SET - Set Bit On
	 *
	 *  @param value		Value to adjust
	 *  @param b			Bit to turn on
	 *
	 *  @return			 Adjusted value
	 */

	private final int setBit(int value, int b){
		b = (0x01<<b);  // Convert 0-7 to real bit
		value |= b;
		return value;
	}


	/**
	 *  CB RES - Reset Bit
	 *
	 *  @param value		Value to adjust
	 *  @param b			Bit to turn off
	 *
	 *  @return			 Adjusted value
	 */

	private final int res(int value, int b){
		b = (0x01<<b);  // Convert 0-7 to real bit
		value &= ~b;
		return value;
	}


	/**
	 *  CB BIT - Test Bit
	 *
	 *  @param value		Value to test
	 *  @param b			Bit of value to test
	 */

	private final void bit(int value, int b){
		b = (0x01<<b);  // Convert 0-7 to real bit
		if ((value & b) != 0){ // BIT IS SET
			F_ &=~ 0x40;
			F_ &=~ 0x04;
			switch (b) {
				case 0x08: F_ |= 0x08; break;
				case 0x20: F_ |= 0x20; break;
				case 0x80: F_ |= 0x80; break;
			}
		} else {
			F_ |= 0x40;
			F_ |= 0x04;
		}
		F_ &=~ 0x02; F_ |= 0x10;
	}



	private final void out(int port, int value) {
		switch (port) {
			case 0xBE:
				first_byte = true;
				if(operation <= 0x02)
					VRAM[location_VRAM] = (byte) value;
				else if (operation == 0x03){
					if (is_sms == 0) { // SMS
						int r, g, b;
						r = (value & 0x3) * 0x55;
						g = ((value >> 2) & 0x3) * 0x55;
						b = ((value >> 4) & 0x3) * 0x55;
						CRAM[location_VRAM & 0x1F] = Graphics.getColorOfRGB(r,g,b);
					} else if (is_sms == 1) { // GG
						int r, g, b;
						r = (value & 0xF) * 0x10;
						g = ((value >> 4) & 0xF) * 0x10;
						b = (/*(value >> 8)*/value & 0xF) * 0x10;
						if ((location_VRAM & 1) == 0) // first byte
							CRAM[(location_VRAM&0x3F)>>1] = Graphics.getColorOfRGB(r,g,0);
						else
							CRAM[((location_VRAM&0x3F)-1)>>1] |= Graphics.getColorOfRGB(0,0,b);
					}
				}
				if (++location_VRAM > 0x3FFF) location_VRAM = 0;
				break;
			case 0xBD:
			case 0xBF:
				if (first_byte) {
					first_byte = false;
					command_byte = value;
				} else {
					first_byte = true;
					if ((value>>4) == 0x08) {
						vdpreg[(value & 0x0F)] = command_byte; // Set reg to previous byte
						if (lineint) {
							if (((value & 0x0F) == 0) && ((command_byte & 0x10) == 0x0))
								irq = false;
							else if (((value & 0x0F) == 0) && ((command_byte & 0x10) == 0x10))
								irq = true;
						}
						if (frameint) {
							if (((value & 0x0F) == 1) && ((command_byte & 0x20) == 0x0))
								irq = false;
							else if (((value & 0x0F) == 1) && ((command_byte & 0x20) == 0x20))
								irq = true;
						}
					} else {
						operation = (value>>6);
						location_VRAM = command_byte + ((value&0x3F)<<8);
						if (operation == 0) {
							read_buffer = (VRAM[location_VRAM]) & 0xFF;
							if (++location_VRAM > 0x3FFF) location_VRAM = 0;
						}
					}
				}
				break;
			case 0x3F:
				if ((value & 0x20) == 0x20) nat1 = 1;
				else nat1 = 0;
				if ((value & 0x80) == 0x80) nat2 = 1;
				else nat2 = 0;
				//if (!europe) {
					nat1 = ~nat1;
					nat2 = ~nat2;
				//}
				break;
			case 0x7E:
			case 0x7F:
				audio.writeReg(value);
				break;
		}
	}



	private final int in(int port) {
		int value = 0;
		switch (port) {
			case 0x00:
				value = gui.readJoyPad(0);
				//if (europe) value |= 0x40;
				/*else*/ value &= ~0x40;
				return value;
			case 0x01:
			case 0x02:
			case 0x03:
			case 0x04:
			case 0x05:
				return 0;
				
			case 0xC0:
			case 0xDC:
				return gui.readJoyPad(1);
			case 0xC1:
			case 0xDD:
				value = gui.readJoyPad(2);
				if (nat1 == 1) value |= 0x40; else value &= ~0x40;
				if (nat2 == 1) value |= 0x80; else value &= ~0x80;
				return value;
			case 0x7E:
			case 0x7F:
				if (line > 0xDA) value = line-6;// Values from 00 to DA, then jump to D5-FF
				else value = line;
				return value;
			case 0xBE:
				first_byte = true; // Reset flag
				if (operation <= 0x02) {
					value = read_buffer;
					read_buffer = (VRAM[location_VRAM]) & 0xFF;
				} else {
					value = 0;
				}
				if (++location_VRAM > 0x3FFF) location_VRAM = 0;
				return value;
			case 0xBD:
			case 0xBF:
				first_byte = true;
				value = status;
				status &=~0x80 &~0x40 &~0x20;
				lineint = false;
				frameint = false;
				irq = false;
				return value;
			default:
				return 0xFF;
		}
	}



	public final void stateSave(DataOutputStream dos) throws IOException {
		dos.write(romCheck);
		dos.writeInt(stateVer);
		dos.writeInt(nat1);
		dos.writeInt(nat2);
		dos.writeInt(status);
		dos.writeInt(command_byte);
		dos.writeInt(location_VRAM);
		dos.writeInt(operation);
		dos.writeInt(read_buffer);
		dos.writeInt(counter);
		dos.writeInt(pc);
		dos.writeInt(sp);
		dos.writeInt(F_);
		dos.writeInt(F2_);
		dos.writeInt(A_);
		dos.writeInt(A2_);
		dos.writeInt(B_);
		dos.writeInt(B2_);
		dos.writeInt(C_);
		dos.writeInt(C2_);
		dos.writeInt(D_);
		dos.writeInt(D2_);
		dos.writeInt(E_);
		dos.writeInt(E2_);
		dos.writeInt(H_);
		dos.writeInt(H2_);
		dos.writeInt(L_);
		dos.writeInt(L2_);
		dos.writeInt(IX_);
		dos.writeInt(IY_);
		dos.writeInt(R_);
		dos.writeInt(I_);
		dos.writeInt(cart_ram_page);
		dos.writeBoolean(pause_button);
		dos.writeBoolean(first_byte);
		dos.writeBoolean(lineint);
		dos.writeBoolean(frameint);
		dos.writeBoolean(iff1);
		dos.writeBoolean(iff2);
		dos.writeBoolean(halt);
		dos.writeBoolean(EI_inst);
		dos.writeBoolean(irq);
		dos.writeBoolean(frame_two_rom);
		dos.write(VRAM);

		for (int i = 0; i < 0x40; i++)	dos.writeInt(CRAM[i]);
		for (int i = 0; i < 0x10; i++)	dos.writeInt(vdpreg[i]);
		for (int i = 0; i < 0x107; i++)	dos.writeBoolean(bg_priority[i]);
		for (int i = 0; i < 0x100; i++)	dos.writeBoolean(spritecol[i]);
		for (int i = 0; i < 2; i++)	dos.write(cart_ram_pages[i]);

		dos.write(mem,0xC000,0x4000);

		dos.writeInt(getIndex(mem,0x0000));
		dos.writeInt(getIndex(mem,0x4000));
		dos.writeInt(getIndex(mem,0x8000));
	}


	private int getIndex(byte b[], int off) {
		byte[] frame = new byte[0x4000];
		System.arraycopy(b,off,frame,0x0000,0x4000);

		boolean flag;
		for(int i=0; i<pages.length; i++) {
			flag = true;
			for(int j=0; j<pages[i].length; j++) {
				if(frame[j] != pages[i][j]) flag = false;
				if(flag) return i;
			}
		}

		for(int i=0; i<cart_ram_pages.length; i++) {
			flag = true;
			for(int j=0; j<cart_ram_pages[i].length; j++) {
				if(frame[j] != cart_ram_pages[i][j]) flag = false;
				if(flag) return i;
			}
		}

		return -1;
	}


	public final void stateLoad(DataInputStream dis) throws IOException {
		byte[] temp = new byte[0x10];
		boolean temp2 = false;

		dis.read(temp);
		for (int i = 0; i < 0x10; i++) {
			if (romCheck[i] != temp[i])
				temp2 = true;
		}
		if (temp2) {
			dis.close();
			dis = null;
			emu_running = true;
			return;
		}
		dummyInt = dis.readInt();
		nat1 = dis.readInt();
		nat2 = dis.readInt();
		status = dis.readInt();
		command_byte = dis.readInt();
		location_VRAM = dis.readInt();
		operation = dis.readInt();
		read_buffer = dis.readInt();
		counter = dis.readInt();
		pc = dis.readInt();
		sp = dis.readInt();
		F_ = dis.readInt();
		F2_ = dis.readInt();
		A_ = dis.readInt();
		A2_ = dis.readInt();
		B_ = dis.readInt();
		B2_ = dis.readInt();
		C_ = dis.readInt();
		C2_ = dis.readInt();
		D_ = dis.readInt();
		D2_ = dis.readInt();
		E_ = dis.readInt();
		E2_ = dis.readInt();
		H_ = dis.readInt();
		H2_ = dis.readInt();
		L_ = dis.readInt();
		L2_ = dis.readInt();
		IX_ = dis.readInt();
		IY_ = dis.readInt();
		R_ = dis.readInt();
		I_ = dis.readInt();
		cart_ram_page = dis.readInt();
		pause_button = dis.readBoolean();
		first_byte = dis.readBoolean();
		lineint = dis.readBoolean();
		frameint = dis.readBoolean();
		iff1 = dis.readBoolean();
		iff2 = dis.readBoolean();
		halt = dis.readBoolean();
		EI_inst = dis.readBoolean();
		irq = dis.readBoolean();
		frame_two_rom = dis.readBoolean();
		dis.read(VRAM);

		for (int i = 0; i < 0x40; i++)	CRAM[i] = dis.readInt();
		for (int i = 0; i < 0x10; i++)	vdpreg[i] = dis.readInt();
		for (int i = 0; i < 0x107; i++)	bg_priority[i] = dis.readBoolean();
		for (int i = 0; i < 0x100; i++)	spritecol[i] = dis.readBoolean();
		for (int i = 0; i < 2; i++)	dis.read(cart_ram_pages[i]);

		dis.read(mem,0xC000,0x4000);
		
		System.arraycopy(pages[dis.readInt()],0x0000,mem,0x0000,0x4000);
		System.arraycopy(pages[dis.readInt()],0x0000,mem,0x4000,0x4000);
		if(frame_two_rom) {
			System.arraycopy(pages[dis.readInt()],0x0000,mem,0x8000,0x4000);
		}
		else {
			int i = dis.readInt();
			if(i<cart_ram_pages.length)
				System.arraycopy(cart_ram_pages[i],0x0000,mem,0x8000,0x4000);
			else
				System.arraycopy(pages[i],0x0000,mem,0x8000,0x4000);
		}
		audio.init();
		emu_running = true;
	}

	public void setRunning(boolean flag){
		emu_running = flag;
	}

	public int[] display(){
		return (is_sms == 0 || gui.getZoom() == 0) ? display : display_zoom;
	}

	public void setSMS(boolean flag){
		if(flag)is_sms = 0;
		else is_sms = 1;
	}

	public void pause_button(boolean flag){
		pause_button = flag;
	}

	public void i_sw(boolean flag){
		i_sw = flag;
	}

	private void drawFrame(){
		for (int raster = 0; raster < 144; raster++) {
			int x = (raster + 24) * 256 + 48;
			int j = (raster >> 1) * 720;
			if ((raster & 1) == 0) {
				for (int i = 0; i < 160; i += 2) {
					int c1 = display[x + i];
					int c2 = display[x + i + 1];
					display_zoom[j] = c1;
					display_zoom[j + 1] = (c1 + c2) >> 1;
					display_zoom[j + 2] = c2;
					j += 3;
				}
			} else {
				for (int i = 0; i < 160; i += 2) {
					int c1 = display[x + i];
					int c2 = display[x + i + 1];
					display_zoom[j + 480] = c1;
					display_zoom[j + 481] = (c1 + c2) >> 1;
					display_zoom[j + 482] = c2;
					display_zoom[j + 240] = (display_zoom[j] + c1) >> 1;
					display_zoom[j + 241] = (display_zoom[j] + c2) >> 1;
					display_zoom[j + 242] = (display_zoom[j + 2] + c2) >> 1;
					j += 3;
				}
			}
		}
	}

}
